# -*- coding: utf-8 -*-
# Copyright © 2020-24 Wacom. All rights reserved.
"""
This package provides the Wacom Ink Service library for Python.
It contains the following sub-packages:

    * nimbus - Nimbus service
    * nel - Named Entity Linking service
    * hwr - Handwriting Recognition service
    * experimental - Experimental services
"""
import logging

__author__ = "Markus Weber"
__copyright__ = "Copyright 2021 Wacom. All rights reserved."
__credits__ = ["Markus Weber"]
__license__ = "Wacom"
__version__ = "1.5.3"
__maintainer__ = ["Markus Weber"]
__email__ = "markus.weber@wacom.com"
__status__ = "beta"

# MIME-type Universal Ink Model data serialization
from typing import Optional, Dict, Any

import aiohttp
from requests import Response

VND_WACOM_INK_MODEL: str = 'application/vnd.wacom-ink.model'
"""MIME-type for Universal Ink Model data serialization."""


class WacomServiceException(Exception):
    """
    Exception class for service related errors.
    """

    def __init__(self, message: str, headers: Optional[Dict[str, Any]] = None, payload: Optional[Dict[str, Any]] = None,
                 params: Optional[Dict[str, Any]] = None, method: Optional[str] = None,
                 url: Optional[str] = None, service_response: Optional[str] = None, status_code: int = 500):
        super().__init__(message)
        self.__status_code: int = status_code
        self.__service_response: Optional[str] = service_response
        self.__message: str = message
        self.__headers: Optional[Dict[str, Any]] = headers
        self.__payload: Optional[Dict[str, Any]] = payload
        self.__params: Optional[Dict[str, Any]] = params
        self.__method: Optional[str] = method
        self.__url: Optional[str] = url

    @property
    def headers(self) -> Optional[Dict[str, Any]]:
        """Headers of the exception."""
        return self.__headers

    @property
    def method(self) -> Optional[str]:
        """Method of the exception."""
        return self.__method

    @property
    def params(self) -> Optional[Dict[str, Any]]:
        """Parameters of the exception."""
        return self.__params

    @property
    def payload(self) -> Optional[Dict[str, Any]]:
        """Payload of the exception."""
        return self.__payload

    @property
    def url(self) -> Optional[str]:
        """URL of the exception."""
        return self.__url

    @property
    def message(self) -> str:
        """Message of the exception."""
        return self.__message

    @property
    def service_response(self) -> Optional[str]:
        """Service response."""
        return self.__service_response

    @property
    def status_code(self) -> int:
        """Status code of the exception."""
        return self.__status_code


def format_exception(exception: WacomServiceException) -> str:
    """
    Formats the exception.

    Parameters
    ----------
    exception: WacomServiceException
        Exception

    Returns
    -------
    formatted_exception: str
        Formatted exception
    """
    return (f'WacomServiceException: {exception.message}\n'
            '--------------------------------------------------\n'
            f'URL:= {exception.url}\n,'
            f'method:= {exception.method}\n,'
            f'parameters:= {exception.params}\n,'
            f'payload:= {exception.payload}\n,'
            f'headers:= {exception.headers}\n,'
            f'status code=: {exception.status_code}\n,'
            f'service response:= {exception.service_response}')


def handle_error(message: str, response: Response, parameters: Optional[Dict[str, Any]] = None,
                 payload: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, str]] = None) \
        -> WacomServiceException:
    """
    Handles an error response.

    Parameters
    ----------
    message: str
        Error message
    response: aiohttp.ClientResponse
        Response
    parameters: Optional[Dict[str, Any]] (Default:= None)
        Parameters
    payload: Optional[Dict[str, Any]] (Default:= None)
        Payload
    headers: Optional[Dict[str, str]] (Default:= None)
        Headers

    Returns
    -------
    WacomServiceException
        Returns the generated exception.
    """
    return WacomServiceException(message,
                                 url=response.url,
                                 method=response.request.method,
                                 params=parameters,
                                 payload=payload,
                                 headers=headers,
                                 status_code=response.status_code,
                                 service_response=response.text)


async def async_handle_error(message: str, response: aiohttp.ClientResponse,
                             parameters: Optional[Dict[str, Any]] = None,
                             payload: Optional[Dict[str, Any]] = None,
                             headers: Optional[Dict[str, str]] = None) -> WacomServiceException:
    """
    Handles an error response.

    Parameters
    ----------
    message: str
        Error message
    response: aiohttp.ClientResponse
        Response
    parameters: Optional[Dict[str, Any]] (Default:= None)
        Parameters
    payload: Optional[Dict[str, Any]] (Default:= None)
        Payload
    headers: Optional[Dict[str, str]] (Default:= None)
        Headers

    Returns
    -------
    WacomServiceException
        Create exception.
    """
    try:
        response_text: str = await response.text()
    except Exception as _:
        response_text: str = ''
    return WacomServiceException(message,
                                 method=response.method,
                                 url=response.url.human_repr(),
                                 params=parameters,
                                 payload=payload,
                                 headers=headers,
                                 status_code=response.status,
                                 service_response=response_text)

DEFAULT_TIMEOUT: int = 60
"""Default timeout for requests."""
# Create the Logger
logger: Optional[logging.Logger] = None

if logger is None:
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    ch: logging.StreamHandler = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    # create formatter and add it to the handlers
    formatter: logging.Formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(ch)
    logger.info('Completed configuring logger()!')

__all__ = ['logger', 'nimbus', 'nel', 'hwr', 'experimental', 'VND_WACOM_INK_MODEL', 'WacomServiceException',
           '__version__', 'DEFAULT_TIMEOUT', 'handle_error', 'async_handle_error', 'format_exception']
from ink.services import nimbus, nel, hwr, experimental
