# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring
# Copyright © 2023-present Wacom. All rights reserved.
import base64
import enum
import hashlib
import json
import random
import time
import uuid
from abc import ABC
from typing import List, Optional, Tuple, Dict, Any, Union

import numpy as np
import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import load_der_public_key
from requests import Response
from uim.model.base import UUIDIdentifier
from uim.model.ink import InkModel, InkTree, logger
from uim.model.inkdata.strokes import Stroke
from uim.model.inkinput.inputdata import InkSensorType, InputContext, SensorContext, SensorChannel
from uim.model.inkinput.sensordata import SensorData, ChannelData
from uim.model.semantics.node import StrokeGroupNode, StrokeNode
from uim.model.semantics.schema import CommonViews, TEXT_REGION, IS, TEXT_LINE, WORD, HAS_CONTENT
from uim.model.semantics.structures import BoundingBox

from ink.services.hwr import REPRESENTS
from ink.services.hwr.myscript import enclosing_bounding_box

GENERATED_BY: str = "generatedBy"
WORDS_TAG: str = 'words'
DATA_TAG: str = 'data'
CODE_TAG: str = 'code'
LABEL_TAG: str = 'label'
CANDIDATES_TAG: str = 'candidates'
BOUNDING_BOX_TAG: str = 'bounding_box'

PUNCTUATION_SYMBOLS: list = [',', '.', '!', ';', '!', '?', '-', ':', '\'', '\"', '-', '(', ')']
SUPPORTED_LANGUAGES: list = ['en', 'de', 'fr', 'jp', 'ja']


class ClientSignatureData:
    """
    Client signature data.
    """
    def __init__(self):
        self.type = "SIMPLE"
        self.publicKeySHA256 = ""
        self.appIdMd5 = ""
        self.emailMd5 = ""
        self.data = ""


class SuniaRecognitionMode(enum.Enum):
    """
    Different mode of recognition for Sunia.
    """
    TEXT = 'rte'
    # Language and text recognition
    MATH_CHEM = 'rme'
    # Mathematics/chemistry formula recognition
    MATH = 'rmc'
    # Mathematics formula recognition
    SHAPE = 'rsp'
    # Shape recognition
    GESTURE = 'rgt'
    # Gesture recognition


class SuniaSaasService(ABC):
    """
    Abstract class for all Sunia services
    """

    TOKEN_ENDPOINT: str = 'aisaas/tokenService/getToken'
    RECOGNITION_ENDPOINT: str = 'aisaas/engineService/engineInvoke/recognize'

    CONCAT_STRING: str = "||"

    def __init__(self, url: str, app_id: Optional[str] = None, token: Optional[str] = None):
        self.__brand_name = "Wacom"
        self.__app_id: str = app_id
        self.__service_url: str = url
        self.__artificial_timestamps: bool = False
        self.__token: Optional[str] = token

    @property
    def service_url(self) -> str:
        """URL of service."""
        return self.__service_url

    @property
    def base_url(self) -> str:
        """Base URL of service."""
        if self.service_url.endswith('/'):
            return f'{self.__service_url}'
        return f'{self.__service_url}/'

    @property
    def app_id(self) -> str:
        """Application id for service."""
        return self.__app_id

    @property
    def token(self) -> str:
        """Token for service."""
        return self.__token

    @property
    def artificial_timestamps(self) -> bool:
        """Artificial timestamps."""
        return self.__artificial_timestamps

    @artificial_timestamps.setter
    def artificial_timestamps(self, value: bool):
        self.__artificial_timestamps = value

    @staticmethod
    def encrypt_data(public_key: str, data: str):
        """
        Encrypt data using public key.

        Parameters
        ----------
        public_key: str
            Public key
        data: str
            Data to encrypt
        """
        public_key = load_der_public_key(base64.b64decode(public_key), backend=default_backend())
        # Encrypt the data
        encrypted_data = public_key.encrypt(
            data.encode(),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        return encrypted_data.hex()

    @staticmethod
    def create_signature(public_key: str, email: str, brand_name: str, app_id: str) -> ClientSignatureData:
        # 1. public key sha256 sign
        salt: str = str(random.randint(0, 999999))  # Use an appropriate range for your use case
        public_key_final = hashlib.sha256(public_key.encode('utf-8')).hexdigest()

        # 2. appId sign
        app_id_temp: str = public_key + SuniaSaasService.CONCAT_STRING + app_id + SuniaSaasService.CONCAT_STRING + salt
        app_id_md5_final = hashlib.md5(app_id_temp.encode('utf-8')).hexdigest()

        # 3. email sign
        email_temp = public_key + SuniaSaasService.CONCAT_STRING + email + SuniaSaasService.CONCAT_STRING + salt
        email_md5_final = hashlib.md5(email_temp.encode('utf-8')).hexdigest()

        # 4. Business data is encrypted using public key (Replace with your encryption method)
        data = (public_key + SuniaSaasService.CONCAT_STRING + app_id + SuniaSaasService.CONCAT_STRING +
                brand_name + SuniaSaasService.CONCAT_STRING + email + SuniaSaasService.CONCAT_STRING + salt)
        data_final = SuniaSaasService.encrypt_data(public_key, data)

        # Assemble the signature object
        client_signature_data = ClientSignatureData()
        client_signature_data.publicKeySHA256 = public_key_final
        client_signature_data.appIdMd5 = app_id_md5_final
        client_signature_data.emailMd5 = email_md5_final
        client_signature_data.data = data_final
        return client_signature_data

    def authenticate(self):
        pass

    @staticmethod
    def bounding_box(bounding_box: dict):
        return BoundingBox(bounding_box['x'], bounding_box['y'], bounding_box['width'], bounding_box['height'])

    @staticmethod
    def __strokes__(ink_doc: InkModel, stroke_ids: List[str]) -> List[dict]:
        """
        Only use all strokes which are selected for recognition. If the list is not set, all strokes will be used.

        Parameters
        ----------
        ink_doc: InkModel
            Ink model
        stroke_ids: List[str]
            List of selected stroke ids

        Returns
        -------
        List of MyScript encoded stroke dicts.
        """

        stroke_arr: List[Dict[str, Any]] = []
        selected_strokes: List[Stroke] = list(filter(lambda stroke: len(stroke_ids) == 0 or stroke.id_s_form
                                                     in stroke_ids or str(stroke.id) in stroke_ids,
                                                     ink_doc.strokes))
        now: int = round(time.time() * 1000) - len(selected_strokes) * 2000
        # Iterate over selected strokes
        for stroke_idx, s in enumerate(selected_strokes):
            # Coordinates
            xs: List[float] = s.splines_x
            ys: List[float] = s.splines_y
            # Remove the initial sample point
            if xs[0] == xs[1] and ys[0] == ys[1]:
                xs = xs[1:-1]
                ys = ys[1:-1]
            # Get timestamp
            ts: List[int] = []
            if s.sensor_data_id:
                sd: SensorData = ink_doc.sensor_data.sensor_data_by_id(s.sensor_data_id)
                if sd is not None:
                    ctx: InputContext = ink_doc.input_configuration.get_input_context(sd.input_context_id)
                    sensor_context: SensorContext = ink_doc.input_configuration \
                        .get_sensor_context(ctx.sensor_context_id)

                    timestamp_channel: SensorChannel = sensor_context.get_channel_by_type(InkSensorType.TIMESTAMP)
                    if timestamp_channel is not None:
                        cd: ChannelData = sd.get_data_by_id(timestamp_channel.id)
                        # convert to milliseconds from default seconds setting
                        ts = [int(t * timestamp_channel.resolution) for t in cd.values.copy()]
            # Artificial timestamps
            if len(ts) == 0:
                ts = [now + (t * 10) for t in range(len(xs))]
                now = ts[-1] + 1000
            # Hacky way of fixing misaligned timestamps with splines
            if len(ts) != len(xs):
                if len(ts) < len(xs):
                    try:
                        avg: int = int(np.mean([ts[i] - ts[i - 1] for i in range(1, len(ts))]))
                    except Exception as e:
                        logger.error(str(e))
                        avg: int = 20
                    for _ in range(len(xs) - len(ts)):
                        ts.append(ts[-1] + avg)
                else:
                    ts = ts[:len(xs)]
            m_points: List[Dict[str, Union[float, int]]] = [{'x': x, 'y': y, 't': t} for x, y, t in zip(xs, ys, ts)]

            stroke: Dict[str, Union[int, List[Any]]] = {
                "mId": stroke_idx,
                "mPoints": m_points
            }
            stroke_arr.append(stroke)
        return stroke_arr

    @classmethod
    def __map_ids__(cls, bounding_box: Dict[str, float], ink_model: InkModel, selected_strokes: List[str])\
            -> List[Stroke]:
        """Mapping the list of items to the stroke ids.

        Parameters
        ----------
        bounding_box: Dict[str, float]
            Bounding box
        ink_model: InkModel
            Ink model
        selected_strokes: List[str]
            List of selected strokes

        Returns
        -------
        list_ids: List[Stroke]
            List of stroke ids.
        """
        strokes: List[Stroke] = []
        x: float = bounding_box['x']
        y: float = bounding_box['y']
        width: float = bounding_box['width']
        height: float = bounding_box['height']
        for stroke in ink_model.strokes:
            if stroke.id_s_form in selected_strokes:
                hit_test: List[bool] = []
                # Check if any point is within the bounding box
                for x_i, y_i in zip(stroke.splines_x, stroke.splines_y):
                    hit_test.append(x <= x_i <= x + width and y <= y_i <= y + height)
                # Check if any hit, if yes, add stroke
                if any(hit_test):
                    strokes.append(stroke)
        return strokes


class SuniaInkToText(SuniaSaasService):
    """
    Sunia Ink to Text service.
    """

    SUPPORTED_LANGUAGES: list = ['en', 'de', 'fr', 'jp', 'ja']
    SUPPORTED_COUNTRIES: list = ['US', 'GB', 'DE', 'FR', 'JP']
    DPI: int = 96  # Device independent pixel 1/96 of inch

    def __init__(self, app_id: Optional[str] = None, token: Optional[str] = None, url: str = 'https://saas.sunia.com'):
        super().__init__(url, app_id, token)

    def supports_language(self, language_code: str):
        """Check if the technology supports the language.

        Parameters
        ----------
        language_code: str
            Language code ISO 639-1

        Returns
        -------
        flag: bool
            Flag that if a language is supported
        """
        return language_code.lower() in self.SUPPORTED_LANGUAGES

    def supports_country(self, country_code: str):
        """Check if the technology supports the country code.

        Parameters
        ----------
        country_code: str
            Country code

        Returns
        -------
        flag: bool
            Flag that if a country is supported
        """
        return country_code.upper() in self.SUPPORTED_COUNTRIES

    @staticmethod
    def build_text_model(ink_object: InkModel, hwr_root: StrokeGroupNode, data_object: Dict[str, Any],
                         selected_stroke_ids: List[str]):
        """
        Build the text model.

        Parameters
        ----------
        ink_object: InkModel
            Ink object
        hwr_root: StrokeGroupNode
            Root node
        data_object: Dict[str, Any]
            Data object
        selected_stroke_ids: List[str]
            List of selected stroke ids
        """
        # --------------------------------------------------------------------------------------------------------------
        text_region: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
        hwr_root.add(text_region)
        ink_object.add_semantic_triple(subject=text_region.uri, predicate=IS, obj=TEXT_REGION)
        # --------------------------------------------------------------------------------------------------------------
        text_line: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
        text_region.add(text_line)
        ink_object.add_semantic_triple(subject=text_line.uri, predicate=IS, obj=TEXT_LINE)
        # --------------------------------------------------------------------------------------------------------------
        # Iterate over recognition units
        if WORDS_TAG in data_object:
            for element in data_object[WORDS_TAG]:
                bounding_box: Dict[str, float] = element[BOUNDING_BOX_TAG]
                # ------------------------------------------------------------------------------------------------------
                word: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
                if BOUNDING_BOX_TAG in element:
                    word.group_bounding_box = SuniaSaasService.bounding_box(element[BOUNDING_BOX_TAG])
                    # --------------------------------------------------------------------------------------------------
                    if LABEL_TAG in element and element[LABEL_TAG] == "\n":
                        # Text line
                        text_line = StrokeGroupNode(UUIDIdentifier.id_generator())
                        text_region.add(text_line)

                        if text_region.group_bounding_box is None:
                            text_region.group_bounding_box = text_line.group_bounding_box
                        else:
                            text_region.group_bounding_box = text_region.group_bounding_box \
                                .enclosing_bounding_box(text_line.group_bounding_box)

                        ink_object.add_semantic_triple(subject=text_line.uri, predicate=IS, obj=TEXT_LINE)
                        continue
                    elif LABEL_TAG in element and element[LABEL_TAG] == " ":
                        pass  # Blank
                    elif LABEL_TAG in element:
                        label: str = element[LABEL_TAG]
                        candidates: List[str] = element.get(CANDIDATES_TAG, [])
                        strokes: List[Stroke] = SuniaInkToText.__map_ids__(bounding_box, ink_object,
                                                                           selected_stroke_ids)
                        for s in strokes:
                            selected_stroke_ids.remove(s.id_s_form)
                        # only add words with associated strokes
                        if len(strokes) > 0:
                            for stroke in strokes:
                                word.add(StrokeNode(stroke))
                                if text_line.group_bounding_box is None:
                                    text_line.group_bounding_box = word.group_bounding_box
                                else:
                                    text_line.group_bounding_box = enclosing_bounding_box(text_line.group_bounding_box,
                                                                                          word.group_bounding_box)

                            text_line.add(word)
                            # ------------------------------------------------------------------------------------------
                            if len(label) == 1 and label in PUNCTUATION_SYMBOLS:
                                ink_object.add_semantic_triple(subject=word.uri, predicate=IS, obj=WORD)
                            else:
                                ink_object.add_semantic_triple(subject=word.uri, predicate=IS, obj=WORD)
                            ink_object.add_semantic_triple(subject=word.uri, predicate=HAS_CONTENT, obj=label)
                            for idx, alt in enumerate(candidates):
                                if len(alt) > 0 and alt != " ":
                                    ink_object.add_semantic_triple(subject=word.uri,  predicate=f"hasAltContent-{idx}",
                                                                   obj=alt)

    @staticmethod
    def build_enriched_model(ink_object: InkModel, results_obj: dict, selected_stroke_ids: List[str]) -> InkModel:
        """Converts the result JSON into a common internal structure.

        Parameters
        ----------
        ink_object: InkModel
            Ink document
        results_obj: dict
            Structure from Ink Analysis service
        selected_stroke_ids: List[str]
            Selected stroke ids
        """
        if not isinstance(ink_object, InkModel):
            raise Exception('Not an Ink Document object!')
        # Create view
        hwr_view: InkTree = InkTree(name=str(CommonViews.HWR_VIEW.value))
        ink_object.add_view(hwr_view)

        data_object: Dict[str, Any] = json.loads(results_obj[DATA_TAG])
        version: str = data_object['version']
        type_obj: str = data_object['type']

        # Adding a root document
        root_id: uuid.UUID = UUIDIdentifier.id_generator()
        hwr_root: StrokeGroupNode = StrokeGroupNode(root_id)
        hwr_view.root = hwr_root
        ink_object.add_semantic_triple(subject=hwr_root.uri, predicate=REPRESENTS, obj='hwr')
        ink_object.add_semantic_triple(subject=hwr_root.uri, predicate=GENERATED_BY, obj=f'sunia (v{version})')
        if type_obj == 'math_generic':
            # Build math model
            raise NotImplementedError("Math recognition is not implemented yet.")
        else:
            # Build text model
            SuniaInkToText.build_text_model(ink_object, hwr_root, data_object, selected_stroke_ids)
        return ink_object

    def process_sync(self, ink_object: InkModel, locale: str = "en_US",
                     content_type: SuniaRecognitionMode = SuniaRecognitionMode.TEXT) \
            -> Tuple[InkModel, Dict[str, Any], List[str]]:
        """Process synchronous request.

        Parameters
        ----------
        ink_object: :class:`uim.model.ink.InkModel` -
            Ink object containing the strokes.
        locale: str
            Locale ISO code
        content_type: ContentType
            Content type
        Returns
        -------
        result: InkModel
            Enriched ink object

        Raises
        ------
        TypeError:
            if ink doc is on InkObject.
        """
        # Start recognition
        stroke_list: List[str] = SuniaInkToText.__parted_request__(ink_object)
        strokes: List[Dict[str, Any]] = SuniaInkToText.__strokes__(ink_object, stroke_list)
        time_stamp: int = int(round(time.time() * 1000))
        code_data: str = f"{self.app_id}%{time_stamp}%{self.token}"
        # Create a SHA-256 hash object
        sha256_hash = hashlib.sha256()
        # Update the hash object with the input data
        sha256_hash.update(code_data.encode('utf-8'))
        # Get the hexadecimal representation of the hash
        verification_code: str = sha256_hash.hexdigest()
        payload: dict = {
            "engineParams": {
                "engineType": content_type.value,
                "subType": locale,
                "engineLines": strokes
            },
            "timeStamp": time_stamp,
            "appId": self.app_id,
            "verificationCode": verification_code
        }
        json_obj: str = json.dumps(payload, ensure_ascii=False)
        url: str = self.base_url + SuniaSaasService.RECOGNITION_ENDPOINT
        headers: dict = {
            "Content-Type": "application/json",
        }
        response_body: Response = requests.post(url, data=json_obj.encode(), headers=headers, timeout=60)
        response_body.encoding = 'utf-8'
        result: dict = {}
        ink_model: InkModel = ink_object
        if response_body.ok:
            result: dict = response_body.json()
            try:
                ink_model: InkModel = self.build_enriched_model(ink_object, result, stroke_list.copy())
            except Exception as e:
                logger.exception(e)
        # Build the model
        return ink_model, result, stroke_list

    @classmethod
    def __parted_request__(cls, ink_object: InkModel) -> List[str]:
        stroke_ids: List[str] = []
        if ink_object.has_tree("parted-request"):
            view: InkTree = ink_object.view("parted-request")
            # Iterate
            for node in view.root.children:
                if isinstance(node, StrokeGroupNode):
                    for s_i in [p for p in node.children if isinstance(p, StrokeNode)]:
                        stroke_ids.append(s_i.id_s_form)
            return stroke_ids
        else:
            return [stroke.id_s_form for stroke in ink_object.strokes]
