# -*- coding: utf-8 -*-
# Copyright © 2022-present Wacom. All rights reserved.
# pylint: disable=missing-module-docstring
import string
from typing import Tuple, List, Dict

from uim.model.helpers.treeiterator import PreOrderEnumerator
from uim.model.ink import InkModel
from uim.model.semantics.node import StrokeNode, InkNode
from uim.model.semantics.schema import CommonViews, WORD, HAS_CONTENT, TEXT_LINE

from ink.services import WacomServiceException

REPRESENTS: str = 'representsView'


def mapping_model_words(ink_model: InkModel, locale: str) -> Tuple[List[dict], str]:
    """
    Mapping model to words.

    Parameters
    ----------
    ink_model: InkModel
        Ink Model
    locale: str
        Locale

    Returns
    -------
    mapping: List[dict]
        List of mappings
    text: str
        Concatenated text
    """
    # All handwriting recognition named_entities are stored within a view:
    # see: https://developer-docs.wacom.com/sdk-for-ink/docs/semantics
    text: str = ''
    mapping: List[dict] = []
    root: InkNode = ink_model.view_root(str(CommonViews.HWR_VIEW.value))
    text_nodes: Dict[str, str] = {}
    text_lines: List[str] = []
    # Iterate for triples with triple list and look for words
    for s in ink_model.knowledge_graph.statements:
        # Collect all words
        if s.object == WORD:
            all_statements = ink_model.knowledge_graph.all_statements_for(s.subject, predicate=HAS_CONTENT)
            if len(all_statements) == 1:
                text_nodes[s.subject] = all_statements[0].object
        # Check for text lines
        elif s.object == TEXT_LINE:
            text_lines.append(s.subject)
    # Position
    pos: int = 0
    # Iterate
    for node in PreOrderEnumerator(root):
        if node.uri in text_lines:
            for word_node in node.children:
                if word_node.uri in text_nodes:
                    t = text_nodes[word_node.uri]
                    path_ids: list = [p.stroke.id for p in word_node.children if isinstance(p, StrokeNode)]
                    start_index = len(text)
                    end_index: int = start_index + len(t)
                    if t in string.punctuation or pos == 0 or locale in ['ja_JP']:
                        text += t
                    else:
                        text += f' {t}'
                        start_index += 1
                        end_index += 1
                    mapping.append({'start_idx': start_index, 'end_idx': end_index, 'text': t, 'position': pos,
                                    'path_ids': path_ids, "uri": word_node.uri})
                    # Position
                    pos += 1
            # text += '\n'
    return mapping, text


class Ink2TextServiceException(WacomServiceException):
    """
    Exception for Ink2Text service.
    """


__all__ = ['REPRESENTS', 'mapping_model_words', 'Ink2TextServiceException']

from ink.services.hwr import msftclient, myscript
