# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring
# Copyright © 2022-24 Wacom. All rights reserved.
import enum
import json
import uuid
from abc import ABC
from typing import Any, Optional, Tuple, List, Dict

import requests
from requests import Response
from uim.model.base import UUIDIdentifier
from uim.model.ink import InkTree, InkModel
from uim.model.inkdata.strokes import Stroke
# List
from uim.model.inkinput.inputdata import unit2unit, Unit
from uim.model.semantics.node import StrokeGroupNode, StrokeNode
from uim.model.semantics.schema import CommonViews, WORD, IS, TEXT_LINE, HAS_CONTENT, TEXT_REGION, ROOT

from ink.services import DEFAULT_TIMEOUT, handle_error
from ink.services.hwr import REPRESENTS

ID: str = 'id'
PARENT: str = 'parent'
CATEGORY: str = 'category'
RECOGNITION_UNITS: str = 'recognitionUnits'
PARAGRAPH: str = 'paragraph'
INK_WORD: str = 'inkWord'
RECOGNIZED_TEXT: str = 'recognizedText'
LINE: str = 'line'
WRITING_REGION: str = 'writingRegion'
LEAF: str = 'leaf'
ALTERNATES: str = 'alternates'
RECOGNIZED_STRING: str = 'recognizedString'
HAS_PARENT: str = 'hasParent'


class InputScope(enum.Enum):
    """
    InputScope
    """
    WORD: str = 'word'
    LINE: str = 'line'
    CHARACTER: str = 'character'
    DEFAULT: str = 'default'
    TEXT: str = 'text'
    URL: str = 'url'
    EMAIL: str = 'email'
    FULL_NAME: str = 'fullName'
    FULL_ADDRESS: str = 'fullAddress'
    POSTALCODE: str = 'postalCode'
    DIGITS: str = 'digits'
    NUMBER: str = 'number'
    TELEPHONE: str = 'telephone'
    DATE: str = 'date'
    MONTH: str = 'month'
    DAY: str = 'day'
    YEAR: str = 'year'
    MONTH_NAME: str = 'monthname'
    DAY_NAME: str = 'dayname'
    TIME: str = 'time'
    HOUR: str = 'hour'
    MINOR_SEC: str = 'minorsec'
    CURRENCY: str = 'currency'
    CURRENCY_WITH_SYMBOL: str = 'currencywithsymbol'
    CURRENCY_WITHOUT_SYMBOL: str = 'currencywithoutsymbol'
    SPREADSHEET: str = 'spreadsheet'


MAP_INPUT_SCOPE: Dict[str, InputScope] = dict([(str(scope.value), scope) for scope in InputScope])


class LineRecognizerService(ABC):
    """
    Ink Recognizer Service
    ----------------------
    Cloud Input Intelligence service for Ink provides JSON based REST APIs for ink analysis and recognition.
    These APIs can be used to extract the layout and content within a set of digital ink strokes.
    The APIs accept ink strokes, uses various machine learning techniques for analyzing them and provide a response
    that can be parsed to understand what the ink strokes contain.
    """
    """Base URL of Microsoft Ink Analysis service."""
    SUPPORTED_LANGUAGES: str = ['en', 'de', 'fr', 'jp']
    SUPPORTED_COUNTRIES: str = ['US', 'GB', 'DE', 'FR', 'JP']
    DEFAULT_CACHE_SIZE: int = 10

    PRODUCTION_URL: str = 'https://input.microsoft.com'

    def __init__(self, subscription_key: str, api_version: str, url: str = PRODUCTION_URL):
        self.__base_service_url = url
        self.__subscription_key = subscription_key
        self.__api_version = api_version

    @property
    def subscription_key(self):
        """Subscription key."""
        return self.__subscription_key

    @property
    def api_version(self):
        """API version."""
        return self.__api_version

    def supports_language(self, language_code: str):
        """Check if the technology supports the language.

        Parameters
        ----------
        language_code: str
            Language code ISO 639-1

        Returns
        -------
        flag: bool
            True if language is supported
        """
        return language_code.lower() in self.SUPPORTED_LANGUAGES

    def supports_country(self, country_code: str):
        """Check if the technology supports the country code.

        Parameters
        ----------
        country_code: str
            Country code

        Returns
        -------
        supported: bool
            True if country is supported
        """
        return country_code.upper() in self.SUPPORTED_COUNTRIES

    @classmethod
    def __strokes__(cls, ink_model: InkModel) -> Tuple[List[dict], Dict[str, str]]:
        """Converts the ink document to predefined structure.

        Parameters
        ----------
        ink_model: InkModel -
            instance of ink document

        Returns
        -------
        stroke_arr: List[dict]
            List of strokes
        """
        stroke_arr: List[dict] = []
        idmap: Dict[str, str] = {}
        id_counter: int = 1
        for s in ink_model.strokes:
            idmap[str(id_counter)] = str(s.id)
            stroke: Dict[str, str] = {
                "id": id_counter,
                "points": LineRecognizerService.__stroke2points__(s),
            }

            stroke_arr.append(stroke)
            id_counter += 1
        return stroke_arr, idmap

    @staticmethod
    def __stroke2points__(stroke: Stroke) -> str:
        """Converts stroke to JSON string.

        Parameters
        ----------
        stroke: Stroke -
            Stroke structure

        Returns
        -------
        stroke_str: str
            Stroke string converted to MM
        """
        xs: List[float] = stroke.splines_x
        ys: List[float] = stroke.splines_y
        points: str = ''
        for x, y in zip(xs[1:], ys[1:]):
            points: str = points + f'{unit2unit(Unit.DIP, Unit.MM, x):.4f},{unit2unit(Unit.DIP, Unit.MM, y):.4f},'
        return points[:-1]

    @staticmethod
    def parameters(ink_model: InkModel, locale: str, input_scope: InputScope) -> Dict[str, Any]:
        """
        Parameters for the request.

        Parameters
        ----------
        ink_model: InkModel
            instance of ink document
        locale: str
            Locale encoding code <language-code>-<country-code>
        input_scope: InputScope
            Scope of the input
        """
        strokes, _ = LineRecognizerService.__strokes__(ink_model)
        payload: dict = {
            "version": 1,
            "language": locale,
            'unit': 'mm',
            "analysisHint": {
                "inputScope": str(input_scope.value)
            },
            'strokes': strokes
        }
        return payload

    @staticmethod
    def enrich_ink(ink_model: InkModel, results_obj: dict, mapping: dict):
        """
        Enriches the ink document with the results from the service.

        Parameters
        ----------
        ink_model: InkModel -
            instance of ink document
        results_obj: dict -
            results from the service
        mapping: dict -
            mapping of strokes from ink document to request
        """
        # Create view
        hwr_view: InkTree = InkTree(str(CommonViews.HWR_VIEW.value))
        ink_model.add_view(hwr_view)
        # --------------------------------------------------------------------------------------------------------------
        # Adding a root document
        root_id: uuid.UUID = UUIDIdentifier.id_generator()
        hwr_root: StrokeGroupNode = StrokeGroupNode(root_id)
        hwr_view.root = hwr_root
        ink_model.add_semantic_triple(subject=hwr_root.uri, predicate=IS, obj=ROOT)
        ink_model.add_semantic_triple(subject=hwr_root.uri, predicate=REPRESENTS, obj='hwr')
        # --------------------------------------------------------------------------------------------------------------
        # Counter for ids
        id_counter: int = 1
        id_mapping: dict = dict([(r['id'], UUIDIdentifier.id_generator()) for r in results_obj[RECOGNITION_UNITS]])
        id_mapping[0] = hwr_view.root.uri
        text_regions: Dict[str, StrokeGroupNode] = {}
        line_nodes: Dict[str, StrokeGroupNode] = {}
        content_node: Optional[StrokeGroupNode] = None
        # Iterate over recognition units
        for unit in sorted(results_obj[RECOGNITION_UNITS], key=lambda x: (x[PARENT], x[ID])):
            id_counter += 1
            cat: str = unit[CATEGORY]
            if cat == PARAGRAPH:
                text_region: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
                text_regions[unit['id']] = text_region
                hwr_root.add(text_region)
                ink_model.add_semantic_triple(subject=text_region.uri, predicate=HAS_PARENT, obj=hwr_root.uri)
                ink_model.add_semantic_triple(subject=text_region.uri, predicate=IS, obj=TEXT_REGION)
            elif cat == LINE:
                # Text line
                line_node: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
                line_nodes[unit['id']] = line_node
                text_regions[unit['parent']].add(line_node)
                ink_model.add_semantic_triple(subject=line_node.uri, predicate=HAS_PARENT,
                                              obj=text_regions[unit['parent']].uri)
                ink_model.add_semantic_triple(subject=line_node.uri, predicate=IS, obj=TEXT_LINE)
            elif cat == INK_WORD:
                content_node = StrokeGroupNode(UUIDIdentifier.id_generator())
                line_nodes[unit['parent']].add(content_node)
                ink_model.add_semantic_triple(subject=content_node.uri, predicate=HAS_PARENT,
                                              obj=line_nodes[unit['parent']].uri)
                ink_model.add_semantic_triple(subject=content_node.uri,
                                              predicate=IS,
                                              obj=WORD)
                if ALTERNATES in unit:
                    for idx, alt in enumerate([a[RECOGNIZED_STRING] for a in unit[ALTERNATES]]):
                        ink_model.add_semantic_triple(subject=content_node.uri,
                                                      predicate=f"hasAltContent-{idx}",
                                                      obj=alt)
                if RECOGNIZED_TEXT in unit:
                    ink_model.add_semantic_triple(subject=content_node.uri,
                                                  predicate=HAS_CONTENT,
                                                  obj=unit[RECOGNIZED_TEXT])
            # Adding the strokes in the tree as STROKE nodes
            if unit['class'] == LEAF and content_node:
                for s_id in [mapping[str(i)] for i in unit["strokeIds"].split(',')]:
                    stroke: Stroke = ink_model.stroke_by_id(uuid.UUID(s_id))
                    content_node.add(StrokeNode(stroke))
        ink_model.calculate_bounds_recursively(hwr_view.root)
        return ink_model

    def process(self, ink_model: InkModel, locale: str = "en-US", input_scope: InputScope = InputScope.TEXT) \
            -> Tuple[InkModel, Dict[str, Any], Dict[str, Any]]:
        """Process handwriting recognition request.

        Parameters
        ----------
        ink_model: InkModel
            instance of ink document
        locale: str
            Locale encoding code <language-code>-<country-code>
        input_scope: InputScope
            Scope of the input

        Returns
        --------
        result: Tuple[InkModel, Dict[str, Any], Dict[str, Any]]
            model: InkModel
                Enriched model
            mapping: Dict[str, Any]
                Mapping for ID of strokes from ink document to request
            operation: Dict[str, Any]
                Id where to pull the content
        """
        if not isinstance(ink_model, InkModel):
            raise Exception('Not an Ink Document object!')

        _, mapping = self.__strokes__(ink_model)
        payload: Dict[str, Any] = LineRecognizerService.parameters(ink_model, locale, input_scope)
        url: str = f"{self.__base_service_url}/Ink/TextRecognition"
        headers: dict = {
            'content-type': 'application/json',
            'Ocp-Apim-Subscription-Key': self.subscription_key,
            'charset': 'utf-8'
        }
        response_body: Response = requests.put(url, data=json.dumps(payload, ensure_ascii=False),
                                               params={'api-version': self.api_version,
                                                       'requestCorrelationId': str(uuid.uuid4())}, headers=headers,
                                               timeout=DEFAULT_TIMEOUT, verify=True)
        if response_body.ok:
            content: dict = response_body.json()
            return self.enrich_ink(ink_model, content, mapping), content, mapping
        raise handle_error("Handwriting recognition failed.", response_body)
