# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring
# Copyright © 2022-present Wacom. All rights reserved.
import datetime
import enum
import sys
import time
import uuid
from abc import ABC
from typing import List, Optional, Tuple, Dict, Any

import numpy as np
import requests
from requests import Response
from uim.model.base import UUIDIdentifier
from uim.model.ink import InkModel, InkTree, logger
from uim.model.inkdata.strokes import Stroke
from uim.model.inkinput.inputdata import InkSensorType, InputContext, SensorContext, SensorChannel
from uim.model.inkinput.sensordata import SensorData, ChannelData
from uim.model.semantics.node import StrokeGroupNode, StrokeNode
from uim.model.semantics.schema import CommonViews, TEXT_REGION, IS, TEXT_LINE, HAS_CONTENT, WORD
from uim.model.semantics.structures import BoundingBox

from ink.services import handle_error, DEFAULT_TIMEOUT
from ink.services.hwr import Ink2TextServiceException, REPRESENTS

CANDIDATES_TAG: str = 'candidates'
ITEMS_TAG: str = 'items'
LABEL_TAG: str = 'label'
WORDS_TAG: str = 'elements'

PUNCTUATION_SYMBOLS: list = [',', '.', '!', ';', '!', '?', '-', ':', '\'', '\"', '-', '(', ')']


def enclosing_bounding_box(bb1: BoundingBox, bb2: BoundingBox) -> BoundingBox:
    """
    Enclosing bounding box.

    Parameters
    ----------
    bb1: BoundingBox
        Bounding box 1
    bb2: BoundingBox
        Bounding box 2

    Returns
    -------
    BoundingBox
        Enclosing bounding box
    """
    new_x: float = min(bb1.x, bb2.x)
    new_y: float = min(bb1.y, bb2.y)
    new_width: float = max(bb1.width, bb1.width + bb1.x - bb2.x)
    new_height: float = max(bb1.height, bb1.height + bb1.y - bb2.y)
    return BoundingBox(new_x, new_y, new_width, new_height)


class ContentType(enum.Enum):
    """Content type for recognition."""
    TEXT = 'Text'
    MATH = 'Math'
    DIAGRAM = 'Diagram'
    RAW_CONTENT = 'Raw Content'


class MyScriptService(ABC):
    """
    Abstract class for all MyScript services
    ----------------------------------------
    This is the base class for all MyScript services.
    """

    def __init__(self, subscription_key: str, api_version: str, url: str = 'https://cloud.myscript.com'):
        self.__subscription_key: str = subscription_key
        self.__api_version: str = api_version
        self.__service_url: str = url
        self.__artificial_timestamps: bool = False

    @property
    def service_url(self) -> str:
        """URL of service."""
        return self.__service_url

    @property
    def subscription_key(self) -> str:
        """Subscription key for service."""
        return self.__subscription_key

    @property
    def api_version(self) -> str:
        """API version"""
        return self.__api_version

    @property
    def artificial_timestamps(self) -> bool:
        """Artificial timestamps."""
        return self.__artificial_timestamps

    @artificial_timestamps.setter
    def artificial_timestamps(self, value: bool):
        self.__artificial_timestamps = value

    @staticmethod
    def bounding_box(bounding_box: dict):
        """Bounding box from MyScript.

        Parameters
        ----------
        bounding_box: dict
            Bounding box dict

        Returns
        -------
        BoundingBox
            Bounding box
        """
        return BoundingBox(bounding_box['x'], bounding_box['y'], bounding_box['width'], bounding_box['height'])

    @staticmethod
    def __strokes__(ink_doc: InkModel, stroke_ids: List[str]) -> Tuple[List[dict], Dict[int, str], float, float]:
        """
        Only use all strokes which are selected for recognition. If the list is not set, all strokes will be used.

        Parameters
        ----------
        ink_doc: InkModel
            Ink model
        stroke_ids: List[str]
            List of selected stroke ids

        Returns
        -------
        List of MyScript encoded stroke dicts.
        """
        height: float = 0.
        width: float = 0.
        stroke_arr: list = []
        idmap: dict = {}
        selected_strokes: List[Stroke] = list(filter(lambda s: len(stroke_ids) == 0 or s.id_s_form in stroke_ids or
                                                     str(s.id) in stroke_ids, ink_doc.strokes))
        now: int = round(time.time() * 1000) - len(selected_strokes) * 2000
        # Iterate over selected strokes
        for s in selected_strokes:
            # Coordinates
            xs: List[float] = s.splines_x
            ys: List[float] = s.splines_y
            # Remove the initial sample point
            if xs[0] == xs[1] and ys[0] == ys[1]:
                xs = xs[1:-1]
                ys = ys[1:-1]
            # Get timestamp
            ts: List[int] = []
            if s.sensor_data_id:
                sd: SensorData = ink_doc.sensor_data.sensor_data_by_id(s.sensor_data_id)
                if sd is not None:
                    ctx: InputContext = ink_doc.input_configuration.get_input_context(sd.input_context_id)
                    sensor_context: SensorContext = ink_doc.input_configuration \
                        .get_sensor_context(ctx.sensor_context_id)

                    timestamp_channel: SensorChannel = sensor_context.get_channel_by_type(InkSensorType.TIMESTAMP)
                    if timestamp_channel is not None:
                        cd: ChannelData = sd.get_data_by_id(timestamp_channel.id)
                        # convert to milliseconds from default seconds setting
                        ts = [int(t * timestamp_channel.resolution) for t in cd.values.copy()]
            # Artificial timestamps
            if len(ts) == 0:
                ts = [now + (t * 10) for t in range(len(xs))]
                now = ts[-1] + 1000
            # Hacky way of fixing misaligned timestamps with splines
            if len(ts) != len(xs):
                if len(ts) < len(xs):
                    try:
                        avg: int = int(np.mean([ts[i] - ts[i - 1] for i in range(1, len(ts))]))
                    except Exception as e:
                        logger.error(str(e))
                        avg: int = 20
                    for _ in range(len(xs) - len(ts)):
                        ts.append(ts[-1] + avg)
                else:
                    ts = ts[:len(xs)]
            width: float = max(width, max(xs))
            height: float = max(height, max(ys))
            idmap[str(ts[0])] = s.id_s_form
            stroke: dict = {
                "x": xs,
                "y": ys,
                "t": ts
            }
            stroke_arr.append(stroke)
        return stroke_arr, idmap, width, height

    @classmethod
    def __map_ids__(cls, items: List[dict], mapping: Dict) -> List[str]:
        """Mapping the list of items to the stroke ids.

        Parameters
        ----------
        items: List[dict] -
            List of items from MyScript result
        mapping: dict -
            List of dict

        Returns
        -------
        list_ids: List[str]
            List of stroke ids.
        """
        ids: List[str] = []
        for i in items:
            date_time_obj: datetime.datetime = datetime.datetime.strptime(i['timestamp'], '%Y-%m-%d %H:%M:%S.%f')
            epoch: datetime.datetime = datetime.datetime(1970, 1, 1)
            diff: datetime.timedelta = date_time_obj - epoch
            elapsed_ms: str = str(int((diff.days * 86400000) + (diff.seconds * 1000) + (diff.microseconds / 1000)))
            if elapsed_ms in mapping:
                ids.append(mapping[elapsed_ms])
            else:
                min_value: int = sys.maxsize
                closed_id: Optional[str] = None
                for v in mapping.keys():
                    diff_int: int = abs(int(elapsed_ms) - int(v))
                    if min_value > diff_int:
                        min_value: int = diff_int
                        closed_id = mapping[v]
                ids.append(closed_id)
        return ids


class InkToText(MyScriptService):
    """MyScripts Interactive Ink Batch Mode service."""

    SUPPORTED_LANGUAGES: list = ['en', 'de', 'fr', 'jp', 'ja']
    SUPPORTED_COUNTRIES: list = ['US', 'GB', 'DE', 'FR', 'JP']
    DPI: int = 96  # Device independent pixel 1/96 of inch

    def __init__(self, subscription_key: str, api_version: str, url: str = 'https://cloud.myscript.com'):
        super().__init__(subscription_key, api_version, url)

    def supports_language(self, language_code: str):
        """Check if the technology supports the language.

        Parameters
        ----------
        language_code: str
            Language code ISO 639-1

        Returns
        -------
        flag: bool
            Flag that if a language is supported
        """
        return language_code.lower() in self.SUPPORTED_LANGUAGES

    def supports_country(self, country_code: str):
        """Check if the technology supports the country code.

        Parameters
        ----------
        country_code: str
            Country code

        Returns
        -------
        flag: bool
            Flag that if a country is supported
        """
        return country_code.upper() in self.SUPPORTED_COUNTRIES

    @staticmethod
    def build_enriched_model(ink_object: InkModel, results_obj: dict, mapping: dict) -> InkModel:
        """Converts the result JSON into a common internal structure.

        Parameters
        ----------
        ink_object: InkModel
            Ink document
        results_obj: dict
            Structure from Ink Analysis service
        mapping: dict
            Mapping of ink ids
        """
        # Create view
        hwr_view: InkTree = InkTree(name=str(CommonViews.HWR_VIEW.value))
        ink_object.add_view(hwr_view)
        # Adding a root document
        root_id: uuid.UUID = UUIDIdentifier.id_generator()
        hwr_root: StrokeGroupNode = StrokeGroupNode(root_id)
        hwr_view.root = hwr_root
        ink_object.add_semantic_triple(subject=hwr_root.uri, predicate=REPRESENTS, obj='hwr')
        stroke_map: Dict[uuid.UUID, Stroke] = dict([(s.id, s) for s in ink_object.strokes])
        # --------------------------------------------------------------------------------------------------------------
        text_region: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
        hwr_root.add(text_region)
        ink_object.add_semantic_triple(subject=text_region.uri, predicate=IS, obj=TEXT_REGION)
        # --------------------------------------------------------------------------------------------------------------
        text_line: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
        text_region.add(text_line)
        ink_object.add_semantic_triple(subject=text_line.uri, predicate=IS, obj=TEXT_LINE)
        # --------------------------------------------------------------------------------------------------------------
        # Iterate over recognition units
        for element in results_obj[WORDS_TAG]:
            # ----------------------------------------------------------------------------------------------------------
            word: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
            if 'bounding-box' in element:
                word.group_bounding_box = InkToText.bounding_box(element['bounding-box'])
            # ----------------------------------------------------------------------------------------------------------
            if LABEL_TAG in element and element[LABEL_TAG] == "\n":
                # Text line
                text_line = StrokeGroupNode(UUIDIdentifier.id_generator())
                text_region.add(text_line)

                if text_region.group_bounding_box is None:
                    text_region.group_bounding_box = text_line.group_bounding_box
                else:
                    text_region.group_bounding_box = text_region.group_bounding_box \
                        .enclosing_bounding_box(text_line.group_bounding_box)

                ink_object.add_semantic_triple(subject=text_line.uri,
                                               predicate=IS,
                                               obj=TEXT_LINE)
                continue
            if LABEL_TAG in element and element[LABEL_TAG] == " ":
                pass  # Blank
            if LABEL_TAG in element and ITEMS_TAG in element:
                ids: List[str] = InkToText.__map_ids__(element[ITEMS_TAG], mapping)
                # only add words with associated strokes
                if len(ids) > 0:
                    for sid in ids:
                        stroke: Stroke = stroke_map[uuid.UUID(sid)]
                        word.add(StrokeNode(stroke))
                        if text_line.group_bounding_box is None:
                            text_line.group_bounding_box = word.group_bounding_box
                        else:
                            text_line.group_bounding_box = enclosing_bounding_box(text_line.group_bounding_box,
                                                                                  word.group_bounding_box)

                    text_line.add(word)
                    # ----------------------------------------------------------------------------------------------
                    if len(element[LABEL_TAG]) == 1 and element[LABEL_TAG] in PUNCTUATION_SYMBOLS:
                        ink_object.add_semantic_triple(subject=word.uri,
                                                       predicate=IS, obj=WORD)
                    else:
                        ink_object.add_semantic_triple(subject=word.uri,
                                                       predicate=IS, obj=WORD)
                    if LABEL_TAG in element:
                        ink_object.add_semantic_triple(subject=word.uri,
                                                       predicate=HAS_CONTENT,
                                                       obj=element[LABEL_TAG])
                    idx: int = 0
                    if CANDIDATES_TAG in element:
                        for alt in element[CANDIDATES_TAG]:
                            ink_object.add_semantic_triple(subject=word.uri,
                                                           predicate=f"hasAltContent-{idx}",
                                                           obj=alt)
                            idx += 1
        return ink_object

    def enqueue_request(self, ink_object: InkModel, locale: str = "en_US",
                        content_type: ContentType = ContentType.TEXT) -> tuple:
        """Performs a handwriting recognition request and adds the results in the document as semantic tags.

        WARNING: Existing semantic tags will be deleted!

        Parameters
        ----------
        ink_object: :class:`uim.model.ink.InkModel`
            instance of ink document
        locale: str
            Language country code
        content_type: ContentType
            Recognition content type

        Returns
        -------
        enriched: InkModel
            Enriched ink model
        """
        stroke_list: List[str] = InkToText.__parted_request__(ink_object)
        strokes, mapping, width, height = InkToText.__strokes__(ink_object, stroke_list)

        payload: dict = {
            "width": width,
            "height": height,
            "xDPI": InkToText.DPI,
            "yDPI": InkToText.DPI,
            "configuration": {
                "lang": locale
            },
            "gesture": {
                "enable": True
            },
            "contentType": content_type.value,
            "strokeGroups": [{"strokes": strokes}]
        }
        url: str = self.service_url + "/api/v" + str(self.api_version) + "/iink/batch"
        headers: dict = {
            "applicationKey": self.subscription_key,
            "Content-Type": "application/json",
            "Accept": "application/vnd.myscript.jiix,application/json"
        }
        response_body: Response = requests.post(url, json=payload, headers=headers, timeout=DEFAULT_TIMEOUT)
        response_body.encoding = 'utf-8'
        result: dict = {}
        if response_body.ok:
            result: dict = response_body.json()
        if response_body.ok:
            return result, mapping
        raise handle_error("Recognition failed.", response_body, headers=headers, payload=payload)

    def process_sync(self, ink_object: InkModel, locale: str = "en_US", content_type: ContentType = ContentType.TEXT) \
            -> Tuple[InkModel, Dict[str, Any], Dict[str, Any]]:
        """Process synchronous request.

        Parameters
        ----------
        ink_object: :class:`uim.model.ink.InkModel` -
            Ink object containing the strokes.
        locale: str
            Locale ISO code
        content_type: ContentType
            Content type
        Returns
        -------
        result: InkModel
            Enriched ink object

        Raises
        ------
        TypeError:
            if ink doc is on InkObject.
        """
        # Start recognition
        response, mapping = self.enqueue_request(ink_object, locale=locale, content_type=content_type)
        try:
            ink_model: InkModel = self.build_enriched_model(ink_object, response, mapping)
        except Exception as e:
            logger.exception(e)
            raise Ink2TextServiceException(str(e)) from e
        # Build the model
        return ink_model, response, mapping

    @classmethod
    def __parted_request__(cls, ink_object: InkModel) -> List[str]:
        """
        Extracts the list of strokes from the ink document.

        Parameters
        ----------
        ink_object: InkModel
            Ink document

        Returns
        -------
        List[str]
            List of stroke ids
        """
        stroke_ids: List[str] = []
        if ink_object.has_tree("parted-request"):
            view: InkTree = ink_object.view("parted-request")
            # Iterate
            for node in view.root.children:
                if isinstance(node, StrokeGroupNode):
                    for s_i in [p for p in node.children if isinstance(p, StrokeNode)]:
                        stroke_ids.append(str(s_i.id))
        return stroke_ids
