# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring
# Copyright © 2022-present Wacom. All rights reserved.
import abc
import string
from typing import *

from knowledge.nel.base import PublicEntityLinkingProcessor, PersonalEntityLinkingProcessor, \
    NamedEntityRecognitionProcessor
from uim.model.helpers.treeiterator import PreOrderEnumerator
from uim.model.ink import InkModel
from uim.model.semantics.node import InkNode, StrokeNode
from uim.model.semantics.schema import CommonViews, WORD, HAS_CONTENT, TEXT_LINE


class NamedEntityUIMProcessor(abc.ABC):
    """
    Named Entity UIM Processor
    --------------------------
    Named Entity Recognition processor for UIM.
    """

    def __init__(self):
        self.__public_entity_linking: Optional[PublicEntityLinkingProcessor] = None
        self.__personal_entity_linking: Optional[PersonalEntityLinkingProcessor] = None
        self.__named_entity_recognition: Optional[NamedEntityRecognitionProcessor] = None

    @property
    def public_entity_linking_engine(self) -> PublicEntityLinkingProcessor:
        """Public entity linking engine."""
        return self.__public_entity_linking

    @public_entity_linking_engine.setter
    def public_entity_linking_engine(self, pel_engine: PublicEntityLinkingProcessor):
        if pel_engine is not None and not isinstance(pel_engine, PublicEntityLinkingProcessor):
            raise TypeError(f'{type(pel_engine)} is not excepted.')
        self.__public_entity_linking = pel_engine

    @property
    def personal_entity_linking_engine(self) -> PersonalEntityLinkingProcessor:
        """Personal entity linking engine."""
        return self.__personal_entity_linking

    @personal_entity_linking_engine.setter
    def personal_entity_linking_engine(self, pel_engine: PersonalEntityLinkingProcessor):
        if engine is not None and not isinstance(pel_engine, PersonalEntityLinkingProcessor):
            raise TypeError(f'{type(pel_engine)} is not excepted.')
        self.__personal_entity_linking = pel_engine

    @property
    def named_entity_recognition_engine(self) -> NamedEntityRecognitionProcessor:
        """Named Entity Recognition engine."""
        return self.__named_entity_recognition

    @named_entity_recognition_engine.setter
    def named_entity_recognition_engine(self, nel_engine: NamedEntityRecognitionProcessor):
        if nel_engine is not None and not isinstance(nel_engine, NamedEntityRecognitionProcessor):
            raise TypeError(f'{type(nel_engine)} is not excepted.')
        self.__named_entity_recognition = nel_engine

    @classmethod
    def __mapping__(cls, ink_model: InkModel) -> Tuple[List[dict], str]:
        """Build a mapping for ink strokes to text.

        Parameters
        ----------
        ink_model: InkModel -
            Universal Ink Model

        Returns
        -------
        result: Tuple[List, str]
            List of mappings and the text concatenated
        """
        # All handwriting recognition named_entities are stored within a view:
        # see: https://developer-docs.wacom.com/sdk-for-ink/docs/semantics
        text: str = ''
        mapping: List[dict] = []
        root: InkNode = ink_model.view_root(str(CommonViews.HWR_VIEW.value))
        text_nodes: Dict[str, str] = {}
        text_lines: List[str] = []
        # Remove old knowledge
        NamedEntityUIMProcessor.__clean__(ink_model)
        # Iterate for triples with triple list and look for words
        for s in ink_model.knowledge_graph.statements:
            # Collect all words
            if s.object == WORD:
                all_statements = ink_model.knowledge_graph.all_statements_for(s.subject, predicate=HAS_CONTENT)
                if len(all_statements) == 1:
                    text_nodes[s.subject] = all_statements[0].object
            # Check for text lines
            elif s.object == TEXT_LINE:
                text_lines.append(s.subject)
        # Position
        pos: int = 0
        # Iterate
        for node in PreOrderEnumerator(root):
            if node.uri in text_lines:
                for word_node in node.children:
                    if word_node.uri in text_nodes:
                        t = text_nodes[word_node.uri]
                        path_ids: list = [p.stroke.id for p in word_node.children if isinstance(p, StrokeNode)]
                        start_index = len(text)
                        end_index: int = start_index + len(t)
                        if t in string.punctuation or pos == 0:
                            text += t
                        else:
                            text += f' {t}'
                            start_index += 1
                            end_index += 1
                        mapping.append({'start_idx': start_index, 'end_idx': end_index, 'text': t, 'position': pos,
                                        'path_ids': path_ids})
                        # Position
                        pos += 1
                # text += '\n'
        return mapping, text

    @staticmethod
    def __clean__(ink_object: InkModel):
        idx: int = 0
        stmt_list: list = ink_object.knowledge_graph.statements
        while idx < len(stmt_list):
            if stmt_list[idx].predicate:
                del stmt_list[idx]
            else:
                idx += 1


__all__ = ['combined', 'engine', 'NamedEntityUIMProcessor', 'PublicEntityLinkingProcessor',
           'PersonalEntityLinkingProcessor', 'NamedEntityRecognitionProcessor']
from ink.services.nel import combined, engine
