# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring
# Copyright © 2022-present Wacom. All rights reserved.
import datetime
import enum
import logging
import uuid
import sys
from typing import List, Tuple, Dict, Any, Optional, Set

from knowledge.base.language import LocaleCode
from knowledge.base.ontology import ThingObject
from knowledge.nel.base import NamedEntity, KnowledgeGraphEntity, BasicNamedEntity, KnowledgeSource, EntityType
from knowledge.public.wikidata import WikiDataAPIClient
from knowledge.services.base import WacomServiceException
from knowledge.services.graph import WacomKnowledgeService
from uim.model.base import UUIDIdentifier
from uim.model.ink import InkModel, InkTree
from uim.model.semantics.node import StrokeGroupNode, URIBuilder, StrokeNode
from uim.model.semantics.schema import SemanticTriple, HAS_PROVIDER, HAS_CREATION_DATE, HAS_ABSTRACT, HAS_IMAGE, \
    CommonViews, HAS_UNIQUE_ID, HAS_URI, HAS_LABEL, HAS_ONTOLOGY_TYPE, HAS_ARTICLE_URL, HAS_TYPE, IS, NAMED_ENTITY, \
    HAS_NAMED_ENTITY, PART_OF_NAMED_ENTITY, HAS_LANGUAGE
from uim.model.semantics.structures import BoundingBox

from ink.services.hwr import mapping_model_words
from ink.services.nel import NamedEntityUIMProcessor


class RecognitionEngine(enum.Enum):
    """
    Recognition Engines for Named Entity Linking.
    """
    PUBLIC_KNOWLEDGE = 0
    PERSONAL_KNOWLEDGE = 1
    ENTITY = 2


class EntityLinkerEngine(NamedEntityUIMProcessor):
    """Generic Entity recognition engine."""

    def __init__(self, service_url: str = 'https://private-knowledge.wacom.com'):
        super().__init__()
        self.__wiki_client: WikiDataAPIClient = WikiDataAPIClient()
        self.__personal_knowledge: WacomKnowledgeService = \
            WacomKnowledgeService(application_name="NEL service", service_url=service_url)

    @staticmethod
    def __words__(start_idx: int, end_idx: int, words: List[Tuple[int, int, str]]) -> List[str]:
        uris: List[str] = []
        for w in words:
            if w[0] >= start_idx and w[1] <= end_idx:
                uris.append(w[2])
        return uris

    @staticmethod
    def __clear__(ink_model: InkModel):
        todo: List[SemanticTriple] = []
        for stmt in ink_model.knowledge_graph:
            if stmt.subject.startswith('uim:ne') or stmt.object.startswith('uim:ne'):
                todo.append(stmt)
        for st in todo:
            ink_model.knowledge_graph.remove_semantic_triple(st)

    @staticmethod
    def __part__(ink_object: InkModel, source: str, predicate: str, target: str, count: int):
        ink_object.add_semantic_triple(subject=source, predicate=predicate, obj=target)
        ink_object.add_semantic_triple(subject=target, predicate=f"hasPart-{count}", obj=source)

    @classmethod
    def __named_entity__(cls, ink_object: InkModel, uri: str, entity: NamedEntity, language: str,
                         entity_map: Dict[str, Dict[str, Any]]):
        curr_dt = datetime.datetime.now()
        timestamp: int = int(round(curr_dt.timestamp()))
        ink_object.add_semantic_triple(subject=uri, predicate=HAS_LANGUAGE, obj=language)
        ink_object.add_semantic_triple(subject=uri, predicate=HAS_PROVIDER, obj='text-razor'
                                       if entity.entity_type == EntityType.PUBLIC_ENTITY else 'wacom')
        ink_object.add_semantic_triple(subject=uri, predicate=HAS_CREATION_DATE, obj=f'{timestamp}')
        if isinstance(entity, KnowledgeGraphEntity):
            content: Optional[Dict[str, Any]] = entity_map.get(entity.entity_source.uri)
            if 'description' in content:
                ink_object.add_semantic_triple(subject=uri, predicate=HAS_ABSTRACT, obj=content['description'])
            if 'image' in content:
                ink_object.add_semantic_triple(subject=uri, predicate=HAS_IMAGE, obj=content['image'])

            if entity.entity_source.source == KnowledgeSource.WIKIDATA:
                ink_object.add_semantic_triple(subject=uri, predicate=HAS_UNIQUE_ID, obj=entity.entity_source.uri)
            elif entity.entity_source.source == KnowledgeSource.WACOM_KNOWLEDGE:
                ink_object.add_semantic_triple(subject=uri, predicate=HAS_URI, obj=entity.entity_source.uri)

            ink_object.add_semantic_triple(subject=uri, predicate=HAS_LABEL, obj=entity.label)
            if entity.entity_source.source == KnowledgeSource.WIKIDATA:
                for o_type in entity.ontology_types:
                    ink_object.add_semantic_triple(subject=uri, predicate=HAS_ONTOLOGY_TYPE, obj=o_type)
            elif entity.entity_source.source == KnowledgeSource.WACOM_KNOWLEDGE and 'type' in content:
                ink_object.add_semantic_triple(subject=uri, predicate=HAS_ONTOLOGY_TYPE, obj=content['type'])
            if entity.content_link is not None:
                ink_object.add_semantic_triple(subject=uri, predicate=HAS_ARTICLE_URL, obj=entity.content_link)

        elif isinstance(entity, BasicNamedEntity):
            ink_object.add_semantic_triple(subject=uri, predicate=HAS_LABEL, obj=entity.ref_text)
            ink_object.add_semantic_triple(subject=uri, predicate=HAS_TYPE, obj=entity.basic_type.name)

    @classmethod
    def __add_strokes__(cls, node_ref: StrokeGroupNode, ink_model: InkModel, path_ids: List[uuid.UUID]):
        x_min: float = sys.float_info.max
        y_min: float = sys.float_info.max
        x_max: float = 0.
        y_max: float = 0.

        # Add strokes to word
        for pid in path_ids:
            node: StrokeNode = StrokeNode(stroke=ink_model.stroke_by_id(pid))
            node_ref.add(node)
            # Bounding box
            x_min = min(x_min, min(node.stroke.splines_x))
            y_min = min(y_min, min(node.stroke.splines_y))
            x_max = max(x_max, max(node.stroke.splines_x))
            y_max = max(y_max, max(node.stroke.splines_y))
        node_ref.group_bounding_box = BoundingBox(x_min, y_min, x_max - x_min, y_max - y_min)

    def __view__(self, ink_object: InkModel, named_entities: List[NamedEntity], mapping: List[dict],
                 language_code: str = 'en_US', user_token: Optional[str] = None):
        uri_builder: URIBuilder = URIBuilder()
        if ink_object.has_tree(str(CommonViews.NER_VIEW.value)):
            ink_object.remove_tree(str(CommonViews.NER_VIEW.value))
        hwr_view: InkTree = ink_object.view(str(CommonViews.HWR_VIEW.value))
        # --------------------------------------------------------------------------------------------------------------
        _, pull_wikidata, pull_personal = EntityLinkerEngine.__prepare_results__(named_entities)
        # already pull all wikidata info's on entities
        entities_wiki_map: Dict[str, Dict[str, Any]] = {}
        if len(pull_wikidata) > 0:
            entities_wiki = self.__wiki_client.retrieve_entities(list(pull_wikidata))
            for e in entities_wiki:
                entities_wiki_map[e.qid] = {
                    'uri': e.qid,
                    'label': e.label_lang(LocaleCode(language_code)).content,
                    'description': e.description_lang(LocaleCode(language_code)).content,
                    'image': e.image(dpi=200),
                    'type': e.instance_of[0].label if len(e.instance_of) > 0 else 'wacom:core#Topic'
                }
        for e in pull_personal:
            try:
                thing: ThingObject = self.__personal_knowledge.entity(e, auth_key=user_token)
                entities_wiki_map[thing.uri] = {
                    'uri': thing.uri,
                    'label': thing.label_lang(LocaleCode(language_code)).content,
                    'description': thing.description_lang(LocaleCode(language_code)).content,
                    'image': thing.image,
                    'type': thing.concept_type.iri
                }
            except WacomServiceException as we:
                logging.exception(we)
        words: List[Tuple[int, int, str]] = []
        # Iterate over mapping
        for i in range(len(mapping)):
            # Current word
            w: dict = mapping[i]
            words.append((w['start_idx'], w['end_idx'], w['uri']))

        # Adding the Named entities
        for e in named_entities:
            ne_uri: str = uri_builder.build_named_entity_uri(UUIDIdentifier.id_generator())
            ink_object.add_semantic_triple(subject=ne_uri, predicate=IS, obj=NAMED_ENTITY)
            ink_object.add_semantic_triple(subject=hwr_view.root.uri, predicate=HAS_NAMED_ENTITY, obj=ne_uri)
            self.__named_entity__(ink_object=ink_object, uri=ne_uri, entity=e, language=language_code,
                                  entity_map=entities_wiki_map)

            count: int = 0
            for u in self.__words__(e.start_idx, e.end_idx, words):
                self.__part__(ink_object, source=u, predicate=PART_OF_NAMED_ENTITY, target=ne_uri, count=count)
                count += 1
        ink_object.calculate_bounds_recursively(hwr_view.root)

    def __add_word__(self, ink_object: InkModel, w_s: Dict[str, Any], word_node_ref: Dict[str, StrokeGroupNode],
                     container: StrokeGroupNode) -> StrokeGroupNode:
        location: str = f'{w_s["start_idx"]}-{w_s["end_idx"]}'
        if location in word_node_ref:
            node: StrokeGroupNode = word_node_ref[location]
        else:
            node: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
            self.__add_strokes__(node, ink_object, w_s['path_ids'])
            word_node_ref[location] = node
            container.add(node)
        return node

    @classmethod
    def __prepare_results__(cls, named_entities: List[NamedEntity]) \
            -> Tuple[Dict[int, List[NamedEntity]], Set[str], Set[str]]:
        pull_wikidata: set = set()
        pull_personal: set = set()

        entities: dict = {}
        # Iterate the entities
        for e in named_entities:
            ignore: bool = False
            if e.start_idx not in entities:
                entities[e.start_idx] = []
            else:
                if isinstance(e, KnowledgeGraphEntity):
                    for e_i in range(len(entities[e.start_idx])):
                        e_c: NamedEntity = entities[e.start_idx][e_i]
                        if isinstance(e_c, KnowledgeGraphEntity):
                            uri: str = e_c.entity_source.uri
                            uri_2: str = e.entity_source.uri
                            if e.entity_type == EntityType.PUBLIC_ENTITY and uri.endswith(uri_2):
                                # In this case public entity has already been re-birthed
                                ignore = True
                                break
            if not ignore:
                entities[e.start_idx].append(e)
            if not ignore and e.entity_type == EntityType.PUBLIC_ENTITY:
                pull_wikidata.add(e.entity_source.uri)
            if not ignore and e.entity_type == EntityType.PERSONAL_ENTITY:
                pull_personal.add(e.entity_source.uri)

        return entities, pull_wikidata, pull_personal

    def nel(self, ink_model: InkModel, user_token: str, locale: LocaleCode) -> Tuple[InkModel, str, List[NamedEntity]]:
        """Performing Named Entity Linking (Recognition).

        Parameters
        ----------
        ink_model: InkModel -
            Universal Ink Model
        user_token: str -
            User token for personal knowledge
        locale: LocaleCode -
            ISO language code

        Returns
        -------
        model: InkModel
            Enriched Universal Ink Model
        """
        mapping, text = mapping_model_words(ink_model, locale)
        entities: List[NamedEntity] = []
        # Named entity recognition
        if self.named_entity_recognition_engine is not None \
                and self.named_entity_recognition_engine.is_language_supported(locale):
            ne = self.named_entity_recognition_engine.named_entities(text, locale)
            entities.extend(ne)
        # Personal knowledge
        if self.personal_entity_linking_engine is not None \
                and self.personal_entity_linking_engine.is_language_supported(locale):
            personal = self.personal_entity_linking_engine.link_personal_entities(text, locale, auth_key=user_token)
            entities.extend(personal)
        # Public knowledge
        if self.public_entity_linking_engine is not None \
                and self.public_entity_linking_engine.is_language_supported(locale):
            public = self.public_entity_linking_engine.link_public_entities(text, locale)
            entities.extend(public)
        EntityLinkerEngine.__clear__(ink_model)
        self.__view__(ink_model, entities, mapping, locale, user_token)
        return ink_model, text, entities
