# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring
# Copyright © 2022-present Wacom. All rights reserved.
from typing import List, Dict

import requests
from knowledge.nel.base import PublicEntityLinkingProcessor, NamedEntityRecognitionProcessor, BasicType, EntityType, \
    NamedEntity, EntitySource, KnowledgeSource, KnowledgeGraphEntity, BasicNamedEntity
from requests import Response


class TextRazorEngine(PublicEntityLinkingProcessor, NamedEntityRecognitionProcessor):
    """
    TextRazor
    ---------
    Public knowledge entity linking.
    """

    TEXT_RAZOR_LANGUAGE: dict = {
        'en_US': 'eng',
        'en_GB': 'eng',
        'zh_CN': 'chi',
        'dn_DK': 'dan',
        'nl_NL': 'dut',
        'fi_FI': 'fin',
        'fr_BE': 'fre',
        'fr_FR': 'fre',
        'fr_CA': 'fre',
        'fr_CH': 'fre',
        'fr_LU': 'fre',
        'de_DE': 'ger',
        'it_IT': 'ita',
        'ja_JP': 'jpn',
        'nb-NO': 'nor',
        'pl-PL': 'pol',
        'pt-BR': 'por',
        'pt-PT': 'por',
        'ru-RU': 'rus',
        'ca-ES': 'spa',
        'sv-SE': 'swe'
    }

    # Mapping of concepts
    ONTOLOGY_TYPES_MAPPING: Dict[str, str] = {
        'SocietalEvent': "wacom:core#SocialEvent",
        'Holiday': "wacom:core#Event",
        'Software': "wacom:business#SoftwareApplication",
        'Person': "wacom:core#Person",
        'Food': "wacom:core#Food",
        'Place': "wacom:core#Location",
        'Website': "wacom:core#Webpage",
        'Organisation': "wacom:core#Organization",
    }

    BASIC_TYPES: Dict[str, BasicType] = {
        'Date': BasicType.DATE,
        'Money': BasicType.MONEY,
        'Person': BasicType.PERSON,
        'Time': BasicType.TIME,
        'Number': BasicType.NUMBER,
        'Place': BasicType.PLACE
    }

    LANGUAGES: List[str] = TEXT_RAZOR_LANGUAGE.keys()

    def __init__(self, api_key: str, service_url: str = 'https://api.textrazor.com/'):
        self.__api_key: str = api_key
        super().__init__(service_url, provider='textrazor', supported_languages=TextRazorEngine.LANGUAGES)

    @classmethod
    def __map_relevant_concept__(cls, e: dict) -> str:
        if 'type' in e:
            types: list = e['type']
            for t in types:
                if t in TextRazorEngine.ONTOLOGY_TYPES_MAPPING:
                    return TextRazorEngine.ONTOLOGY_TYPES_MAPPING[t]
        return 'wacom:core#Thing'

    @classmethod
    def __map_basic_type__(cls, types: List[str]) -> BasicType:
        for t in types:
            if t in TextRazorEngine.BASIC_TYPES:
                return TextRazorEngine.BASIC_TYPES[t]
        return BasicType.UNKNOWN

    @classmethod
    def __map_types__(cls, e: dict) -> list:
        types: list = []
        if 'type' in e:
            types = [TextRazorEngine.ONTOLOGY_TYPES_MAPPING[t] for t in e['type']
                     if t in TextRazorEngine.ONTOLOGY_TYPES_MAPPING]
        if len(types) == 0:
            return ['wacom:core#Topic']
        return types

    def process(self, text: str, language: str, entity_type: EntityType) -> List[NamedEntity]:
        """Processing text.

        Parameters
        ----------
        text: str -
            Text to perform Named entity recognition and entity linking.
        language: str -
            Language code
        entity_type: EntityType -
            Only pull certain entity type

        Returns
        -------
        named_entities: List[NamedEntity]
            List of named entities
        """
        named_entities: List[NamedEntity] = []
        extractors: str = 'entities'
        payload: dict = {
            'text': text,
            'extractors': extractors,
            'languageOverride': self.TEXT_RAZOR_LANGUAGE[language]
        }
        headers: dict = {
            'x-textrazor-key': self.__api_key,
            'Content-Type': 'application/x-www-form-urlencoded'
        }

        response: Response = requests.post(self.service_url, headers=headers, data=payload)
        if response.ok:
            max_confidence: float = 1.
            results: dict = response.json()
            # Iterate over results
            if 'entities' in results['response']:
                for e in results['response']['entities']:
                    entity_types: List[str] = []
                    # --------------------------- Entity content -------------------------------------------------------
                    if entity_type == EntityType.PUBLIC_ENTITY and 'wikidataId' in e:
                        source: EntitySource = EntitySource(e['wikidataId'], KnowledgeSource.WIKIDATA)
                        content_url: str = e['wikiLink'] if 'wikiLink' in e else ''
                        # --------------------------- Ontology types ---------------------------------------------------
                        if 'type' in e:
                            entity_types.extend(e['type'])
                        # ----------------------------------------------------------------------------------------------
                        ne: KnowledgeGraphEntity = KnowledgeGraphEntity(ref_text=e['matchedText'],
                                                                        start_idx=e['startingPos'],
                                                                        end_idx=e['endingPos'],
                                                                        label=e['entityId'],
                                                                        confidence=e['confidenceScore'],
                                                                        source=source, content_link=content_url,
                                                                        ontology_types=entity_types,
                                                                        entity_type=entity_type)
                        ne.relevant_type = TextRazorEngine.relevant_type(ne.ontology_types)
                        named_entities.append(ne)
                        if max_confidence < e['confidenceScore']:
                            # Use max confidence value to normalize
                            max_confidence: float = e['confidenceScore']

                    if entity_type == EntityType.NAMED_ENTITY and 'wikidataId' not in e:
                        basic_entity: BasicType = TextRazorEngine.__map_basic_type__(e['type'])
                        entity: BasicNamedEntity = BasicNamedEntity(ref_text=e['matchedText'],
                                                                    start_idx=e['startingPos'],
                                                                    end_idx=e['endingPos'],
                                                                    basic_type=basic_entity)
                        named_entities.append(entity)
            # Normalize the confidence value
            if entity_type == EntityType.PUBLIC_ENTITY:
                for e in named_entities:
                    if isinstance(e, KnowledgeGraphEntity):
                        e.confidence /= max_confidence
        return named_entities

    def named_entities(self, text: str, language: str = 'en_US') -> List[NamedEntity]:
        return self.process(text, language, EntityType.NAMED_ENTITY)

    def link_public_entities(self, text: str, language: str = 'en_US') \
            -> List[NamedEntity]:
        return self.process(text, language, EntityType.PUBLIC_ENTITY)

    @classmethod
    def relevant_type(cls, entity_type: List[str]) -> str:
        for e in entity_type:
            if e in TextRazorEngine.ONTOLOGY_TYPES_MAPPING:
                return TextRazorEngine.ONTOLOGY_TYPES_MAPPING[e]
        return 'wacom:core#Topic'
