# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring
# Copyright © 2022-present Wacom. All rights reserved.
import json
from abc import ABC
from enum import Enum
from http import HTTPStatus
from typing import Dict

import requests
from requests import Response

from ink.services import handle_error, DEFAULT_TIMEOUT


class WacomServiceAPIClient(ABC):
    """
    Wacom Service API Client
    ------------------------
    Service API client for Wacom services.

    """
    PRODUCTION_URL: str = 'https://cloud.wacom.com'
    STAGING_URL: str = 'https://stage-cloud.wacom.com'
    AUTH_ENDPOINT: str = 'auth/user'
    USER_ENDPOINT: str = 'user'
    USER_LOGIN_ENDPOINT: str = f'{USER_ENDPOINT}/login'

    def __init__(self, application_name: str, service_url: str = PRODUCTION_URL):
        self.__service_url: str = service_url
        self.__application_name: str = application_name
        self.__verify_calls: bool = True

    def auth_token(self, identification_token: str, tenant: str, region: str = 'EU'):
        """
        Token

        Parameters
        -----------
        identification_token: str -
            Identification token.
        tenant: str -
            Tenant key
        region: str -
            Region

        Returns
        -------
        tokens: Tuple[str, str]
            main_token: str
                JWT token
            refresh_token:str
                Refresh token
        """
        url: str = f'{self.service_url}/{WacomServiceAPIClient.AUTH_ENDPOINT}'

        payload: dict = {
            'userId': identification_token,
            'profile': {
                "email": ""
            },
            'region': region,
            'isActive': True
        }

        headers: Dict[str, str] = {
            'x-tenant-api-key': tenant,
            'Content-Type': 'application/json'
        }

        response: Response = requests.patch(url, headers=headers, data=json.dumps(payload))
        result: dict = response.json()
        if response.ok:
            return result['mainToken'], result['refreshToken']
        raise handle_error('Authentication failed.', response, headers=headers, payload=payload)

    def login_user(self, tenant_key: str, external_id: str) -> str:
        """Login as user by using the tenant key and its external user id.

        Parameters
        ----------
         tenant_key: str -
            Tenant key
         external_id: str -
            External id.

        Returns
        -------
        token: str
            Token
        """
        url: str = '{}/{}/'.format(self.service_url, WacomServiceAPIClient.USER_LOGIN_ENDPOINT)
        headers: dict = {
            'x-tenant-api-key': tenant_key
        }
        payload: dict = {
            'externalUserId': external_id
        }
        response: Response = requests.post(url, headers=headers, json=payload, verify=self.verify_calls,
                                           timeout=DEFAULT_TIMEOUT)
        if response.ok:
            return response.text
        raise handle_error('Login failed.', response, headers=headers, payload=payload)

    @property
    def service_url(self):
        """Service URL."""
        return self.__service_url

    @property
    def application_name(self):
        """Application name."""
        return self.__application_name

    @property
    def verify_calls(self):
        """Certificate verification activated."""
        return self.__verify_calls

    @verify_calls.setter
    def verify_calls(self, value: bool):
        self.__verify_calls = value


class Priority(Enum):
    """
    Priority
    """
    LOWEST: str = 'Lowest'
    LOW: str = 'Low'
    NORMAL: str = 'Normal'
    HIGH: str = 'High'
    HIGHEST: str = 'Highest'


class HWRMode(Enum):
    """
    Handwriting recognition modes.
    - Text - Pure text recognition mode
    - Diagram - Engine recognises shapes and text
    - Math - Math content is recognised
    """
    TEXT_MODE: str = 'Text'
    DIAGRAM_MODE: str = 'TextAndShapes'


class Schema(Enum):
    """"
    Schema for representing the results of the recognition.
    """
    SEGMENTATION_V03: str = 'seg_0.3'
    SIMPLE_SEGMENTATION_V01: str = 'smpl-seg/0.1'
    MATH_STRUCTURES_V01: str = 'math-structures_0.1'
    MATH_V06: str = 'math_0.6'


class ExportFormat(Enum):
    """
    Export format.
        - SVG,
        - PNG,
        - JPG,
        - PSD
    """
    SVG = 'SVG'
    PNG = 'PNG'
    JPG = 'JPG'
    PSD = 'PSD'


class PDFType(Enum):
    """
    PDF Export type: Vector and Raster
    """
    RASTER = 'raster'
    VECTOR = 'vector'


class Provider(Enum):
    """
    Provider of recognition.
    """
    MYSCRIPT = 'myscript'
    ILABO = 'ilabo'


class NimbusServicesAPIClient(WacomServiceAPIClient):
    """
    Nimbus Services API Client
    ===========================
    Nimbus service engineering.

    Parameters
    ----------
    application_name: str
        Application name
    service_url: str
        Service URL
    base_url: str
        Base URL for the services
    """
    NER_ENDPOINT: str = '/ner/enrich-uim'
    NER_DOC_ENDPOINT: str = '/ner/enrich-wdf'
    HWR_MODEL_ENDPOINT: str = '/ink-to-text/enrich-uim'
    HWR_DOCUMENT_ENDPOINT: str = '/ink-to-text/enrich-wdf'
    MATH_MODEL_ENDPOINT: str = '/ink-to-math/enrich-uim'
    MATH_DOCUMENT_ENDPOINT: str = '/ink-to-math/enrich-wdf'
    EXPORT_ENDPOINT: str = '/conversion/export-uim'
    PDF_ENDPOINT: str = '/conversion/uim-to-pdf'
    JSON_ENDPOINT: str = '/conversion/uim-to-json'
    WILL_ENDPOINT: str = '/conversion/will2-to-uim'
    WDF_ENDPOINT: str = '/conversion/uim-to-wdf'

    def __init__(self, application_name: str, service_url: str = WacomServiceAPIClient.PRODUCTION_URL,
                 base_url: str = '/v1/exports'):
        super().__init__(application_name, service_url)
        self.__base_url: str = base_url

    @property
    def service_base_url(self):
        """Service base URL."""
        return f'{self.service_url}{"" if self.service_url.endswith("/") else "/"}{self.__base_url}'

    def ner_model(self, content: bytes, user_token: str, locale: str = 'en_US') -> bytes:
        """NER service for Universal Ink Model.

        Parameter
        ---------
        content: bytes
            UIM bytes.
        user_token: str
            Nimbus authentication token.
        locale: str
            Locale.

        Returns
        -------
        enriched_model: InkModel
            Enriched Universal Ink Model.
        """
        headers: Dict[str, str] = {
            'Authorization': f'Bearer {user_token}',
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_base_url}/{NimbusServicesAPIClient.NER_ENDPOINT}'
        response: Response = requests.post(url, headers=headers, data=content, params={'locale': locale},
                                           timeout=DEFAULT_TIMEOUT)
        if response.status_code == HTTPStatus.OK:
            return response.content
        raise handle_error('Service returns failure.', response, headers=headers)

    def ner_document(self, content: bytes, user_token: str, locale: str = 'en_US') -> bytes:
        """NER service for Wacom Document.

        Parameter
        ---------
        content: Wacom document bytes
        user_token: Nimbus authentication token
        locale: locale
        ner_engine: NER engine used for named entity recognition

        Returns
        -------
        document: bytes
            Enriched Wacom Document
        """
        headers: Dict[str, str] = {
            'Authorization': f'Bearer {user_token}',
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_url}{NimbusServicesAPIClient.NER_DOC_ENDPOINT}'
        response: Response = requests.post(url, headers=headers, data=content, params={'locale': locale},
                                           timeout=DEFAULT_TIMEOUT)
        if response.status_code == HTTPStatus.OK:
            return response.content
        raise handle_error('Service returns failure.', response, headers=headers)

    def ink_to_math(self, content: bytes, user_token: str,  schema: Schema,
                    provider: Provider = Provider.MYSCRIPT,
                    priority: Priority = Priority.LOWEST):
        """
        Run math recognition process.

        Parameters
        ----------
        content: bytes
            UIM content
        user_token: str
            User token
        schema: Schema
            Schema for representing the results
        provider: Provider
            Technology provider
        priority: Priority (default:= LOWEST)
            Priority of request

        Returns
        -------
        content: bytes
            Content of enriched model.
        """
        headers: Dict[str, str] = {
            'Authorization': f'Bearer {user_token}',
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_base_url}{NimbusServicesAPIClient.MATH_MODEL_ENDPOINT}'
        parameters: Dict[str, str] = {
            'provider': provider.value,
            'schema': schema.value,
            'priority': priority.value
        }
        response: Response = requests.post(url, headers=headers, data=content, params=parameters,
                                           timeout=DEFAULT_TIMEOUT)
        if response.status_code == HTTPStatus.OK:
            return response.content
        raise handle_error('Ink to Math service failed', response, headers=headers, parameters=parameters)

    def ink_to_math_document(self, content: bytes, user_token: str, schema: Schema,
                             provider: Provider = Provider.MYSCRIPT,
                             priority: Priority = Priority.LOWEST):
        """
        Run math recognition process.

        Parameters
        ----------
        content: bytes
            UIM content
        user_token: str
            User token
        schema: Schema
            Schema for representing the results
        provider: Provider
            Technology provider
        priority: Priority (default:= LOWEST)
            Priority of request

        Returns
        -------
        content: bytes
            Content of enriched model.
        """
        headers: Dict[str, str] = {
            'Authorization': f'Bearer {user_token}',
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_base_url}{NimbusServicesAPIClient.MATH_MODEL_ENDPOINT}'
        parameters: Dict[str, str] = {
            'provider': provider.value,
            'schema': schema.value,
            'priority': priority.value
        }
        response = requests.post(url, headers=headers, data=content, params=parameters)
        if response.status_code == HTTPStatus.OK:
            return response.content
        if response.status_code == HTTPStatus.OK:
            return response.content
        raise handle_error('Service returns failure.', response, headers=headers, parameters=parameters)

    def ink_to_text(self, content: bytes, user_token: str, locale: str = 'en_US', hwr_mode: HWRMode = HWRMode.TEXT_MODE,
                    priority: Priority = Priority.LOWEST, provider: Provider = Provider.MYSCRIPT,
                    schema: Schema = Schema.SEGMENTATION_V03) -> bytes:
        """
        Run handwriting recognition process.

        Parameters
        ----------
        content: bytes
            UIM content
        user_token: str
            User token
        locale: str
            Locale for language and country
        hwr_mode: HWRMode (default:= TEXT_MODE)
            Handwriting recognition mode.
        priority: Priority (default:= LOWEST)
            Priority of request
        provider: str
            Provider of handwriting recognition.
        schema: Schema
            The UIM schema for representing the results
        Returns
        -------
        content: bytes
            Content of enriched model.
        """
        headers: Dict[str, str] = {
            'Authorization': f'Bearer {user_token}',
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_base_url}{NimbusServicesAPIClient.HWR_MODEL_ENDPOINT}'
        parameters: Dict[str, str] = {
            'textprovider': provider.value,
            'locale': locale,
            'mode': hwr_mode.value,
            'priority': priority.value,
            'schema': schema.value
        }
        response: Response = requests.post(url, headers=headers, data=content, params=parameters,
                                           timeout=DEFAULT_TIMEOUT)
        if response.ok:
            return response.content
        raise handle_error('Service returns failure.', response, headers=headers, parameters=parameters)

    def ink_to_text_document(self, content: bytes, user_token: str, locale: str = 'en_US',
                             hwr_mode: HWRMode = HWRMode.TEXT_MODE, priority: Priority = Priority.LOWEST) -> bytes:
        """
        Run handwriting recognition process.

        Parameters
        ----------
        content: bytes
            UIM content
        user_token: str
            User token
        locale: str
            Locale for language and country
        hwr_mode: HWRMode (default:= TEXT_MODE)
            Handwriting recognition mode.
        priority: Priority (default:= LOWEST)
            Priority of request

        Returns
        -------
        content: bytes
            Content of enriched Wacom Document.
        """
        headers: Dict[str, str] = {
            'Authorization': f'Bearer {user_token}',
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_base_url}{NimbusServicesAPIClient.HWR_MODEL_ENDPOINT}'
        parameters: Dict[str, str] = {
            'locale': locale,
            'mode': hwr_mode.value,
            'priority': priority.value
        }
        response: Response = requests.post(url, headers=headers, data=content, params=parameters,
                                           timeout=DEFAULT_TIMEOUT)
        if response.status_code == HTTPStatus.OK:
            return response.content
        raise handle_error('Service returns failure.', response, headers=headers, parameters=parameters)

    def export(self, content: bytes, user_token: str, export: ExportFormat) -> bytes:
        """
        Export service.

        Parameters
        ----------
        content: bytes
            Universal Ink Model
        user_token: str
            Token within Nimbus
        export: ExportFormat
            Export image format

        Returns
        -------
        content: bytes
            Byte content of the enriched UIM model
        """
        headers: Dict[str, str] = {
            'Authorization': f'Bearer {user_token}',
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_base_url}{NimbusServicesAPIClient.EXPORT_ENDPOINT}'
        parameters: Dict[str, str] = {
            'format': export.value
        }
        response: Response = requests.post(url, headers=headers, data=content, params=parameters)
        if response.ok:
            return response.content
        raise handle_error('Export service returns failure.', response, headers=headers, parameters=parameters)

    def export_pdf(self, content: bytes, user_token: str, pdf_type: PDFType):
        """Export service.

        Parameters
        ----------
        content: bytes
            Universal Ink Model
        user_token: str
            Token within Nimbus
        pdf_type: PDFType
            Raster or Vector mode

        Returns
        -------
        pdf_content: bytes
            File content of the result PDF
        """
        headers: Dict[str, str] = {
            'Authorization': 'Bearer {}'.format(user_token),
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_base_url}{NimbusServicesAPIClient.PDF_ENDPOINT}'
        response: Response = requests.post(url, headers=headers, data=content, params={'type': pdf_type.value})
        if response.ok:
            return response.content
        raise handle_error('Export service returns failure.', response, headers=headers,
                           parameters={'type': pdf_type.value})

    def export_json(self, content: bytes, user_token: str):
        """Export service UIM to JSON representation.

        Parameters
        ----------
        content: bytes
            Universal Ink Model
        user_token: str
            Token within Nimbus

        Returns
        -------
        json_content: bytes
            File content of the result JSON
        """
        headers: Dict[str, str] = {
            'Authorization': f'Bearer {user_token}',
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_base_url}{NimbusServicesAPIClient.JSON_ENDPOINT}'
        response: Response = requests.post(url, headers=headers, data=content)
        if response.ok:
            return response.content
        raise handle_error('Export service returns failure.', response, headers=headers)

    def import_will2(self, content: bytes, user_token: str):
        """Import WILL2.

        Parameters
        ----------
        content: bytes
            Universal Ink Model
        user_token: str
            Token within Nimbus

        Returns
        -------
        uim_content: bytes
            File content of UIM
        """
        headers: Dict[str, str] = {
            'Authorization': f'Bearer {user_token}',
            'Content-Type': 'application/octet-stream'
        }
        url: str = f'{self.service_base_url}{NimbusServicesAPIClient.WILL_ENDPOINT}'
        response: Response = requests.post(url, headers=headers, data=content)
        if response.ok:
            return response.content
        raise handle_error('Export service returns failure.', response, headers=headers)
