# -*- coding: utf-8 -*-
# Copyright © 2023-present Wacom. All rights reserved.
from typing import Union, Literal, Any, Dict, Optional

# Commands for entity sourcing
REMOVE_DOCUMENT_COMMAND: str = 'REMOVE_DOCUMENT'
ADD_DOCUMENT_COMMAND: str = 'ADD_DOCUMENT'
UPDATE_DOCUMENT_COMMAND: str = 'UPDATE_DOCUMENT'
BATCH_DOCUMENT_COMMAND: str = 'BATCH_DOCUMENT'

# Commands for entity sourcing
REMOVE_WORD_COMMAND: str = 'REMOVE_WORD'
ADD_WORD_COMMAND: str = 'ADD_WORD'
BATCH_WORD_COMMAND: str = 'BATCH_WORD'


class DocumentModel:
    """
    DocumentModel
    =============
    Request model for building the vector database index.

    Properties
    ----------
    uri: str
        URI of document.
    metadata: Dict[str, Any]
        Metadata filters.
    locale: str
        ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>', e.g., en_US.
    content: str
        Content of the entity
    """

    def __init__(self, uri: str, content: str, locale: str, metadata: Dict[str, Any]):
        self.__uri: str = uri
        self.__content: str = content
        self.__locale: str = locale
        self.__metadata: Dict[str, Any] = metadata

    @property
    def uri(self) -> str:
        """URI of the entity."""
        return self.__uri

    @property
    def content(self) -> str:
        """ Content of the entity."""
        return self.__content

    @property
    def locale(self) -> str:
        """ ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>', e.g., en_US."""
        return self.__locale

    @property
    def metadata(self) -> Dict[str, Any]:
        """Metadata."""
        return self.__metadata

    def __dict__(self):
        return {
            'uri': self.uri,
            'content': self.content,
            'locale': self.locale,
            'metadata': self.metadata
        }


class UpdateDocumentModel:
    """
    UpdateDocumentModel
    ===================
    Request model for updating the vector database index.
    """

    def __init__(self, uri: str, metadata: Dict[str, Any], locale: str, content: Optional[str] = None):
        self.__uri: str = uri
        self.__metadata = metadata
        self.__locale = locale
        self.__content = content

    @property
    def uri(self) -> str:
        """URI of the entity."""
        return self.__uri

    @property
    def metadata(self) -> Dict[str, Any]:
        """Metadata."""
        return self.__metadata

    @property
    def locale(self) -> str:
        """ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>', e.g., en_US."""
        return self.__locale

    @property
    def content(self) -> Optional[str]:
        """Content of the entity."""
        return self.__content

    def __dict__(self):
        return {
            'uri': self.uri,
            'metadata': self.metadata,
            'locale': self.locale,
            'content': self.content
        }


class DeleteDocumentModel:
    """
    DeleteDocumentModel
    ===================
    Request model for deleting the vector database index.

    Properties
    ----------
    uri: str
        URI of the entity
    locale: str
        ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>', e.g., en_US.
    """

    def __init__(self, uri: str, locale: str):
        self.__uri = uri
        self.__locale = locale

    @property
    def uri(self) -> str:
        """URI of the entity."""
        return self.__uri

    @property
    def locale(self) -> str:
        """ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>', e.g., en_US."""
        return self.__locale

    def __dict__(self):
        return {
            'uri': self.uri,
            'locale': self.locale
        }


class OperationModel:
    """
    OperationModel
    ==========
    Request model for NER service

    Properties
    ----------
    kind: Literal[ADD_DOCUMENT_COMMAND, REMOVE_DOCUMENT_COMMAND, UPDATE_DOCUMENT_COMMAND]
        Type of the operation. Either ADD_DOCUMENT_COMMAND or REMOVE_DOCUMENT_COMMAND.
    document: Union[DocumentModel, UpdateDocumentModel, DeleteDocumentModel]
        Entity to be added, updated or removed.

    """

    def __init__(self, kind: Literal[ADD_DOCUMENT_COMMAND, REMOVE_DOCUMENT_COMMAND, UPDATE_DOCUMENT_COMMAND],
                 document: Union[DocumentModel, UpdateDocumentModel, DeleteDocumentModel]):
        self.__kind = kind
        self.__document = document

    @property
    def kind(self) -> Literal[ADD_DOCUMENT_COMMAND, REMOVE_DOCUMENT_COMMAND, UPDATE_DOCUMENT_COMMAND]:
        """Type of the operation. Either ADD_DOCUMENT_COMMAND or REMOVE_DOCUMENT_COMMAND."""
        return self.__kind

    @property
    def document(self) -> Union[DocumentModel, UpdateDocumentModel, DeleteDocumentModel]:
        """Entity to be added, updated or removed."""
        return self.__document

    def __dict__(self):
        return {
            'kind': self.kind,
            'document': self.document.__dict__()
        }


class WordModel:
    """
    DocumentModel
    =============
    Request model for building the vector database index.
    """

    def __init__(self, uri: str, label: str, locale: str, metadata: Dict[str, Any]):
        self.__uri = uri
        self.__label = label
        self.__locale = locale
        self.__metadata = metadata

    @property
    def uri(self) -> str:
        """URI of the entity."""
        return self.__uri

    @property
    def label(self) -> str:
        """ Content of the entity."""
        return self.__label

    @property
    def locale(self) -> str:
        """ ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>', e.g., en_US."""
        return self.__locale

    @property
    def metadata(self) -> Dict[str, Any]:
        """Metadata."""
        return self.__metadata

    def __dict__(self):
        return {
            'uri': self.uri,
            'label': self.label,
            'locale': self.locale,
            'metadata': self.metadata
        }


class DeleteWordModel:
    """
    DeleteWordModel
    ===============
    Request model for deleting the vector database index.
    """

    def __init__(self, uri: str, locale: str):
        self.__uri = uri
        self.__locale = locale

    @property
    def uri(self) -> str:
        """URI of the entity."""
        return self.__uri

    @property
    def locale(self) -> str:
        """ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>', e.g., en_US."""
        return self.__locale

    def __dict__(self):
        return {
            'uri': self.uri,
            'locale': self.locale
        }


class WordOperationModel:
    """
    WordOperationModel
    ==========
    Request model for NER service
    """

    def __init__(self, kind: Literal[ADD_WORD_COMMAND, REMOVE_WORD_COMMAND], word: Union[WordModel, DeleteWordModel]):
        self.__kind = kind
        self.__word = word

    @property
    def kind(self) -> Literal[ADD_WORD_COMMAND, REMOVE_WORD_COMMAND]:
        """Type of the operation. Either ADD_DOCUMENT_COMMAND or REMOVE_DOCUMENT_COMMAND."""
        return self.__kind

    @property
    def word(self) -> Union[WordModel, DeleteWordModel]:
        """Word to be added or removed."""
        return self.__word

    def __dict__(self):
        return {
            'kind': self.kind,
            'word': self.word.__dict__()
        }
