# -*- coding: utf-8 -*-
# Copyright © 2023-present Wacom. All rights reserved.
import socket
import ssl
from typing import Dict, Any, Optional, List

import aiohttp
import orjson
import requests
from aiohttp import ClientTimeout
from knowledge.services.asyncio.base import cached_getaddrinfo
from requests import Response
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from ink.services import handle_error, DEFAULT_TIMEOUT, WacomServiceException
from ink.services.vector import IndexMode
from ink.services.vector.model.index import OperationModel, WordOperationModel


class CachedResolver(aiohttp.resolver.AbstractResolver):
    """
    CachedResolver
    ==============
    Cached resolver for aiohttp.
    """

    async def close(self) -> None:
        pass

    async def resolve(self, host: str, port: int = 0, family: int = socket.AF_INET):
        infos = await cached_getaddrinfo(host)
        return [{'hostname': host, 'host': info[4][0], 'port': port, 'family': family, 'proto': 0,
                 'flags': socket.AI_NUMERICHOST} for info in infos]


cached_resolver: CachedResolver = CachedResolver()
""" Cached resolver for aiohttp."""


async def async_handle_error(message: str, response: aiohttp.ClientResponse,
                             parameters: Optional[Dict[str, Any]] = None, payload: Optional[Dict[str, Any]] = None,
                             headers: Optional[Dict[str, str]] = None) -> WacomServiceException:
    """
    Handles an error response.

    Parameters
    ----------
    message: str
        Error message
    response: aiohttp.ClientResponse
        Response
    parameters: Optional[Dict[str, Any]] (Default:= None)
        Parameters
    payload: Optional[Dict[str, Any]] (Default:= None)
        Payload
    headers: Optional[Dict[str, str]] (Default:= None)
        Headers

    Returns
    -------
    WacomServiceException
        Create exception.
    """
    try:
        response_text: str = await response.text()
    except Exception as _:
        response_text: str = ''
    return WacomServiceException(message,
                                 method=response.method,
                                 url=response.url.human_repr(),
                                 params=parameters,
                                 payload=payload,
                                 headers=headers,
                                 status_code=response.status,
                                 service_response=response_text)


class SemanticIndexClient:
    """
    SemanticSearchClient
    ====================
    Named Entity Linking test client.

    Parameters
    ----------
    service_url: str
        Service URL for the client.
    base_endpoint: str
        Base endpoint for the client.
    verify_calls: bool
        Verify calls.
    """
    def __init__(self, service_url: str = "http://localhost:8000", base_endpoint: str = "vector/api/v1",
                 verify_calls: bool = False):
        self.__service_url: str = service_url
        self.__base_endpoint: str = base_endpoint
        self.__verify_calls: bool = verify_calls

    @property
    def base_url(self) -> str:
        """
        Base URL for the client.
        """
        return f'{self.__service_url}/{self.__base_endpoint}/'

    @staticmethod
    def __async_session__() -> aiohttp.ClientSession:
        """
        Returns an asynchronous session.

        Returns
        -------
        session: aiohttp.ClientSession
            Asynchronous session
        """
        timeout: ClientTimeout = ClientTimeout(total=60)
        ssl_context: ssl.SSLContext = ssl.create_default_context()
        connector: aiohttp.TCPConnector = aiohttp.TCPConnector(ssl=ssl_context, limit_per_host=10,
                                                               resolver=cached_resolver)
        return aiohttp.ClientSession(
            json_serialize=lambda x: orjson.dumps(x).decode(),
            timeout=timeout,
            connector=connector
        )

    def build_documents_index(self, tenant_id: str, operations: List[OperationModel],
                              max_retries: int = 100, backoff_factor: float = 0.5):
        """
        Build index for documents.

        Parameters
        ----------
        tenant_id: str
            Tenant ID
        operations: List[OperationModel]
            Operations
        max_retries: int
            Maximum number of retries
        backoff_factor: float
            Backoff factor

        Raises
        ------
        WacomServiceException
            If the request fails.
        """
        url: str = f"{self.base_url}index/documents/"
        headers: Dict[str, str] = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        mount_point: str = 'https://' if url.startswith('https') else 'http://'
        with requests.Session() as session:
            retries: Retry = Retry(total=max_retries,
                                   backoff_factor=backoff_factor,
                                   status_forcelist=[502, 503, 504])
            session.mount(mount_point, HTTPAdapter(max_retries=retries))
            response: Response = session.post(url, headers=headers,
                                              json={
                                                    "tenantId": tenant_id,
                                                    "operations": [o.__dict__() for o in operations]
                                              },
                                              timeout=DEFAULT_TIMEOUT,
                                              verify=self.__verify_calls)
            if not response.ok:
                raise handle_error("Failed the building document index request.", response, headers=headers,
                                   payload={
                                       "tenantId": tenant_id,
                                       "operations": operations
                                 })

    async def abuild_documents_index(self, tenant_id: str, operations: List[OperationModel]):
        """
        Build index for documents.

        Parameters
        ----------
        tenant_id: str
            Tenant ID
        operations: List[OperationModel]
            Operations

        Raises
        ------
        WacomServiceException
            If the request fails.
        """
        url: str = f"{self.base_url}index/documents/"
        headers: Dict[str, str] = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        async with self.__async_session__() as session:
            async with session.post(url, headers=headers, json={
                                                    "tenantId": tenant_id,
                                                    "operations": [o.__dict__() for o in operations]
                                              },
                                    timeout=DEFAULT_TIMEOUT) as response:
                if not response.ok:
                    raise await async_handle_error("Failed the building document index request.", response,
                                                   headers=headers, payload={
                                                        "tenantId": tenant_id,
                                                        "operations": operations
                                                   })

    def build_labels_index(self, tenant_id: str, operations: List[WordOperationModel],
                           max_retries: int = 100, backoff_factor: float = 0.5):
        """
        Build word index.

        Parameters
        ----------
        tenant_id: str
            Tenant ID
        operations: List[WordOperationModel]
            Operations
        max_retries: int
            Maximum number of retries
        backoff_factor: float
            Backoff factor
        Returns
        -------
        response_dict: Dict[str, Any]
            result from service

        Raises
        ------
        WacomServiceException
            If the request fails.
        """
        url: str = f"{self.base_url}index/labels/"
        headers: Dict[str, str] = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        mount_point: str = 'https://' if url.startswith('https') else 'http://'
        with requests.Session() as session:
            retries: Retry = Retry(total=max_retries,
                                   backoff_factor=backoff_factor,
                                   status_forcelist=[502, 503, 504])
            session.mount(mount_point, HTTPAdapter(max_retries=retries))
            response: Response = session.post(url, headers=headers, json={
                                                    "tenantId": tenant_id,
                                                    "operations": [o.__dict__() for o in operations]
                                              },
                                              timeout=DEFAULT_TIMEOUT,
                                              verify=self.__verify_calls)
            if not response.ok:
                raise handle_error("Failed the building labels index request.", response, headers=headers,
                                   payload={
                                            "tenantId": tenant_id,
                                            "operations": operations
                                      })

    async def abuild_labels_index(self, tenant_id: str, operations: List[WordOperationModel]):
        """
        Build word index.

        Parameters
        ----------
        tenant_id: str
            Tenant ID
        operations: List[WordOperationModel]
            Operations

        Returns
        -------
        response_dict: Dict[str, Any]
            result from service

        Raises
        ------
        WacomServiceException
            If the request fails.
        """
        url: str = f"{self.base_url}index/labels/"
        headers: Dict[str, str] = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        async with self.__async_session__() as session:
            async with session.post(url, headers=headers, json={
                                                    "tenantId": tenant_id,
                                                    "operations": [o.__dict__() for o in operations]
                                              },
                                    timeout=DEFAULT_TIMEOUT) as response:
                if not response.ok:
                    raise await async_handle_error("Failed the building labels index request.", response,
                                                   headers=headers, payload={
                                                        "tenantId": tenant_id,
                                                        "operations": operations
                                                   })

    def delete_index(self, tenant: str, locale: str, filters: Optional[Dict[str, Any]] = None,
                     index_mode: IndexMode = IndexMode.DOCUMENT_INDEX, max_retries: int = 100,
                     backoff_factor: float = 0.5):
        """
        Delete index for a tenant.

        Parameters
        ----------
        tenant: str
            Tenant ID
        locale: str
            Locale
        index_mode: IndexMode
            Index mode
        filters: Optional[Dict[str, Any]] (Default:= None)
            Filter for the index
        max_retries: int
            Maximum number of retries
        backoff_factor: float
            Backoff factor
        """

        url: str = f"{self.base_url}index/"
        headers: Dict[str, str] = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        filter_settings: Dict[str, Any] = filters if filters else {}
        mount_point: str = 'https://' if url.startswith('https') else 'http://'
        with requests.Session() as session:
            retries: Retry = Retry(total=max_retries,
                                   backoff_factor=backoff_factor,
                                   status_forcelist=[502, 503, 504])
            session.mount(mount_point, HTTPAdapter(max_retries=retries))
            response: Response = session.delete(url, headers=headers, json={
                                                                "tenant": tenant,
                                                                "locale": locale,
                                                                "index": index_mode.value,
                                                                "filter": filter_settings
                                                            }, timeout=DEFAULT_TIMEOUT,
                                                verify=self.__verify_calls)
            if not response.ok:
                raise handle_error("Deleting documents failed.", response, headers=headers,
                                   payload={
                                     "tenant": tenant,
                                     "locale": locale,
                                     "index": index_mode.value,
                                     "filter": filter_settings
                                   })

    async def adelete_index(self, tenant: str, locale: str, filters: Optional[Dict[str, Any]] = None,
                            index_mode: IndexMode = IndexMode.DOCUMENT_INDEX):
        """
        Delete index for a tenant.

        Parameters
        ----------
        tenant: str
            Tenant ID
        locale: str
            Locale
        index_mode: IndexMode
            Index mode
        filters: Optional[Dict[str, Any]] (Default:= None)
            Filter for the index
        """

        url: str = f"{self.base_url}index/"
        headers: Dict[str, str] = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        filter_settings: Dict[str, Any] = filters if filters else {}
        async with SemanticIndexClient.__async_session__() as session:
            async with session.delete(url, headers=headers, json={
                                                                "tenant": tenant,
                                                                "locale": locale,
                                                                "index": index_mode.value,
                                                                "filter": filter_settings
                                                            }, timeout=DEFAULT_TIMEOUT) as response:
                if not response.ok:
                    raise await async_handle_error("Deleting documents failed.", response, headers=headers,
                                                   payload={
                                                     "tenant": tenant,
                                                     "locale": locale,
                                                     "index": index_mode.value,
                                                     "filter": filter_settings
                                                   })
